package com.ibm.swg.app.jobplan;

import java.rmi.RemoteException;

import javax.script.ScriptException;

import com.ibm.swg.util.DJPConstants;
import com.ibm.swg.util.PreparedScript;

import psdi.mbo.MboRemote;
import psdi.mbo.MboSetRemote;
import psdi.mbo.MboValue;
import psdi.mbo.MboValueAdapter;
import psdi.mbo.Translate;
import psdi.server.MXServer;
import psdi.util.MXApplicationException;
import psdi.util.MXException;

public class FldDJPJobCalcExpression extends MboValueAdapter implements DJPConstants {

	
	public FldDJPJobCalcExpression(MboValue mbv) {
		super(mbv);
	}

	@Override
	public void init() throws MXException, RemoteException {
		super.init();
		
		MboRemote jobplan = mboValue.getMbo();	// could be a JOBTASK, JLOBLABOR, JOBMATERIAL, JOBSERVICE, JOBTOOL, WPLABOR, WOACTIVITY, WPLABOR, WPMATERIAL, WPSERVICE, WPTOOL
		if ( !jobplan.isBasedOn("JOBPLAN") ) {
			jobplan = jobplan.getOwner();
			if ( !jobplan.isBasedOn("JOBPLAN") ) {
				mboValue.setFlag(READONLY, true);
				return;
			}
		}
		
		
		Translate t = getTranslator();
		String calcMethod = t.toInternalString("CALCMETHOD", getMboValue(DJPConstants.CALCMETHOD).getString());
		boolean byAttribute = DJPConstants.CALCMETHOD_PROPORTIONAL_ATTRIBUTE.equals(calcMethod) || DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE.equals(calcMethod);
		
		mboValue.setFlag(READONLY, jobplan.getMboValueData("JPDURATION").isReadOnly() || !byAttribute);
		mboValue.setFlag(REQUIRED, byAttribute);
	}
	
	@Override
	public void validate() throws MXException, RemoteException {

		String scriptAsText = mboValue.getString();
		PreparedScript<Double> script = null;
		try {
			script = new PreparedScript<Double>(scriptAsText);
		} catch (IllegalArgumentException e ) {
			throw new MXApplicationException("dynamicjobplan","invalidExpression",e);
		}
		
		//script.setParameterValue(HOURS_TAG, 1.0);
		//script.setParameterValue(QUANTITY_TAG, 1.0);
		
		MboSetRemote assetAttributes = MXServer.getMXServer().getMboSet("ASSETATTRIBUTE", mboValue.getMbo().getUserInfo());
		MboSetRemote featureAttributes = MXServer.getMXServer().getMboSet("ASSETFEATURESPEC", mboValue.getMbo().getUserInfo());
		
		featureAttributes.setQbeExactMatch(true);
		assetAttributes.setQbeExactMatch(true);
		
		for(String param : script.getParametersName()) {
			if ( script.getParameter(param) == null ) {
				
				if ( script.paramIsFeature(param)) {
					featureAttributes.resetQbe();
					featureAttributes.setQbe("FEATURE", param.split(":")[0]);
					featureAttributes.setQbe("ASSETATTRID", param.split(":")[1]);
					if ( featureAttributes.count() != 0 ) {
						script.setParameterValue(param, 1.0);
					} else {
						throw new MXApplicationException("dynamicjobplan","invalidAttribute",new Object[]{param});
					}
				} else {
					assetAttributes.resetQbe();
					assetAttributes.setQbe("ASSETATTRID",param);
					if ( assetAttributes.count() != 0 ) {
						script.setParameterValue(param, 1.0);
					} else {
						throw new MXApplicationException("dynamicjobplan","invalidAttribute",new Object[]{param});
					}
				}
				
			}
		}
		assetAttributes.close();
		
		try {
			script.eval();
//			Class<?> clazz = res.getClass();
//			if ( !(	clazz.equals(Double.class) || 
//					clazz.equals(Integer.class) || 
//					clazz.equals(Long.class) ||
//					clazz.equals(Short.class) ||
//					clazz.equals(Float.class) ||
//					clazz.equals(BigDecimal.class))) {
//				throw new MXApplicationException("dynamicjobplan","invalidResultType", new Object[]{clazz});
//			}
		} catch (ScriptException e1) {
			e1.printStackTrace();
			throw new MXApplicationException("dynamicjobplan","invalidExpression", e1);
		}
		
	}
	
	
}
