package com.ibm.swg.app.jobplan;

import java.rmi.RemoteException;

import psdi.mbo.MboRemote;
import psdi.mbo.MboValue;
import psdi.mbo.SynonymDomain;
import psdi.mbo.Translate;
import psdi.util.MXException;

import com.ibm.swg.util.DJPConstants;

public class FldDJPJobMaterialCalcMethod extends SynonymDomain implements DJPConstants {

	public FldDJPJobMaterialCalcMethod(MboValue mv) {
		super(mv);
		setDomainId("CALCMETHOD");
	}
	
	@Override
	public void init() throws MXException, RemoteException {
		super.init();
		
		MboRemote jobplan = mboValue.getMbo().getOwner();
		if ( !jobplan.isBasedOn("JOBPLAN") ) {
			mboValue.setFlag(READONLY, true);
			return;
		}
		
		mboValue.setFlag(REQUIRED, true);
		mboValue.setFlag(READONLY, jobplan.getMboValueData("JPDURATION").isReadOnly() );
		
		String calcMethod = getTranslator().toInternalString("CALCMETHOD", mboValue.getString());
		MboValue itemQty = getMboValue("ITEMQTY");
		
		if ( DJPConstants.CALCMETHOD_LEVEL.equals(calcMethod) || DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE.equals(calcMethod) ) {
			itemQty.setFlag(READONLY, true);	
		} else {
			itemQty.setFlag(READONLY, false);
		}
	}
	
	@Override
	public void action() throws MXException, RemoteException {
		super.action();
		MboValue calcExpression 		= getMboValue(DJPConstants.CALCEXPRESSION);
		MboValue calcExpressionHours 	= getMboValue(DJPConstants.CALCEXPRESSIONHOURS);
		MboValue calcLevel      		= getMboValue(DJPConstants.CALCLEVEL);
		MboValue calcApplyTo			= getMboValue(DJPConstants.CALCAPPLYTO);
		MboValue itemQty				= getMboValue("ITEMQTY");
		
		Translate t = getTranslator();
		String calcMethod = t.toInternalString("CALCMETHOD", mboValue.getString());
		
		String applyTo = t.toInternalString("CALCAPPLYTO", calcApplyTo.getString());
		boolean isBoth = DJPConstants.CalcApplyTo.BOTH.toString().equals(applyTo);
		
		calcLevel.setValueNull(NOACCESSCHECK);
		calcExpression.setValueNull(NOACCESSCHECK);
		calcExpressionHours.setValueNull(NOACCESSCHECK);
		
		calcExpression.setFlag(REQUIRED, DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE.equals(calcMethod) || DJPConstants.CALCMETHOD_PROPORTIONAL_ATTRIBUTE.equals(calcMethod));
		calcExpression.setFlag(READONLY, DJPConstants.CALCMETHOD_STATIC.equals(calcMethod) || DJPConstants.CALCMETHOD_PROPORTIONAL.equals(calcMethod) || DJPConstants.CALCMETHOD_LEVEL.equals(calcMethod));
		
		calcExpressionHours.setFlag(REQUIRED, isBoth && (DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE.equals(calcMethod) || DJPConstants.CALCMETHOD_PROPORTIONAL_ATTRIBUTE.equals(calcMethod)));
		calcExpressionHours.setFlag(READONLY, !isBoth ||  (DJPConstants.CALCMETHOD_STATIC.equals(calcMethod) || DJPConstants.CALCMETHOD_PROPORTIONAL.equals(calcMethod) || DJPConstants.CALCMETHOD_LEVEL.equals(calcMethod)));
		
		calcLevel.setFlag(REQUIRED, DJPConstants.CALCMETHOD_LEVEL.equals(calcMethod) || DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE.equals(calcMethod));
		calcLevel.setFlag(READONLY, DJPConstants.CALCMETHOD_STATIC.equals(calcMethod) || DJPConstants.CALCMETHOD_PROPORTIONAL.equals(calcMethod) || DJPConstants.CALCMETHOD_PROPORTIONAL_ATTRIBUTE.equals(calcMethod));
		
		// as calcApplyTo strictly = QUANTITY
		if ( DJPConstants.CALCMETHOD_LEVEL.equals(calcMethod) || DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE.equals(calcMethod) ) {
			itemQty.setValue(1.0, NOACCESSCHECK);
			itemQty.setFlag(READONLY, true);	
		} else {
			itemQty.setFlag(READONLY, false);
		}
		
		
		calcApplyTo.setFlag(READONLY, DJPConstants.CALCMETHOD_STATIC.equals(calcMethod));
	}

}
