package com.ibm.swg.app.jobplan;

import java.rmi.RemoteException;

import psdi.mbo.MboConstants;
import psdi.mbo.MboRemote;
import psdi.mbo.MboValue;
import psdi.mbo.SynonymDomain;
import psdi.mbo.Translate;
import psdi.util.MXException;

import com.ibm.swg.util.DJPConstants;

public class FldDJPJobToolCalcApplyTo extends SynonymDomain implements DJPConstants {

	public FldDJPJobToolCalcApplyTo(MboValue mv) {
		super(mv);
		setDomainId("CALCAPPLYTO");
	}
	
	
	@Override
	public void init() throws MXException, RemoteException {
		super.init();
		
		MboRemote jobplan = mboValue.getMbo().getOwner();
		if ( !jobplan.isBasedOn("JOBPLAN") ) {
			mboValue.setFlag(READONLY, true);
			return;
		}
			
		String calcMethod = getTranslator().toInternalString("CALCMETHOD", getMboValue(DJPConstants.CALCMETHOD).getString());
		mboValue.setFlag(REQUIRED, jobplan.getBoolean(DJPConstants.DYNAMIC));
		mboValue.setFlag(READONLY, jobplan.getMboValueData("JPDURATION").isReadOnly() || DJPConstants.CALCMETHOD_STATIC.equals(calcMethod));
		
		if ( DJPConstants.CALCMETHOD_LEVEL.equalsIgnoreCase(calcMethod) || DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE.equalsIgnoreCase(calcMethod)) {
			MboValue hours = getMboValue("HOURS");
			MboValue quantity = getMboValue("ITEMQTY");
			
			String calcApplyTo = getTranslator().toInternalString("CALCAPPLYTO", getMboValue(DJPConstants.CALCAPPLYTO).getString());
			switch(CalcApplyTo.fromStringTo(calcApplyTo)) {
			case HOURS:
				hours.setFlag(MboConstants.READONLY, true);
				quantity.setFlag(MboConstants.READONLY, false);
				break;
			case QUANTITY:
				quantity.setFlag(MboConstants.READONLY, true);
				hours.setFlag(MboConstants.READONLY, false);
				break;
			case BOTH:
				hours.setFlag(MboConstants.READONLY, true);
				quantity.setFlag(MboConstants.READONLY, true);
				break;
			}
		}
	}

	@Override
	public void action() throws MXException, RemoteException {
		super.action();
		Translate t = getTranslator();
		String calcMethod = t.toInternalString("CALCMETHOD", getMboValue(DJPConstants.CALCMETHOD).getString());
		if ( mboValue.getMbo().isFlagSet(MboConstants.READONLY) ) {
			return;
		}
		
		if ( DJPConstants.CALCMETHOD_LEVEL.equalsIgnoreCase(calcMethod) || DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE.equalsIgnoreCase(calcMethod)) {
			MboValue toolHours = getMboValue("HOURS");
			MboValue quantity = getMboValue("ITEMQTY");
			
			String calcApplyTo = t.toInternalString("CALCAPPLYTO", getMboValue(DJPConstants.CALCAPPLYTO).getString());
			switch(CalcApplyTo.fromStringTo(calcApplyTo)) {
			case HOURS:
				toolHours.setValue(1.0, MboConstants.NOACCESSCHECK);
				toolHours.setFlag(MboConstants.READONLY, true);
				quantity.setFlag(MboConstants.READONLY, false);
				break;
			case QUANTITY:
				quantity.setValue(1.0, MboConstants.NOACCESSCHECK);
				quantity.setFlag(MboConstants.READONLY, true);
				toolHours.setFlag(MboConstants.READONLY, false);
				break;
			case BOTH:
				toolHours.setValue(1.0, MboConstants.NOACCESSCHECK);
				toolHours.setFlag(MboConstants.READONLY, true);
				quantity.setValue(1.0, MboConstants.NOACCESSCHECK);
				quantity.setFlag(MboConstants.READONLY, true);
				break;
			}
			
		}
	}
}
