package com.ibm.swg.app.jobplan;

import java.rmi.RemoteException;

import psdi.mbo.MboConstants;
import psdi.mbo.MboRemote;
import psdi.mbo.MboValue;
import psdi.mbo.SynonymDomain;
import psdi.mbo.Translate;
import psdi.util.MXException;

import com.ibm.swg.util.DJPConstants;

public class FldDJPJobToolCalcMethod extends SynonymDomain implements DJPConstants {

	public FldDJPJobToolCalcMethod(MboValue mv) {
		super(mv);
		setDomainId("CALCMETHOD");
	}
	
	@Override
	public void init() throws MXException, RemoteException {
		super.init();
		
		MboRemote jobplan = mboValue.getMbo().getOwner();
		if ( !jobplan.isBasedOn("JOBPLAN") ) {
			mboValue.setFlag(READONLY, true);
			return;
		}
		
		mboValue.setFlag(REQUIRED, true);
		mboValue.setFlag(READONLY, jobplan.getMboValueData("JPDURATION").isReadOnly() );
		
		Translate t = getTranslator();
		String calcMethod = t.toInternalString("CALCMETHOD", mboValue.getString());
		if ( DJPConstants.CALCMETHOD_LEVEL.equalsIgnoreCase(calcMethod) || DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE.equalsIgnoreCase(calcMethod)) {
			MboValue hours = getMboValue("HOURS");
			MboValue quantity = getMboValue("ITEMQTY");
			
			String calcApplyTo = t.toInternalString("CALCAPPLYTO", getMboValue(DJPConstants.CALCAPPLYTO).getString());
			switch(CalcApplyTo.fromStringTo(calcApplyTo)) {
			case HOURS:
				hours.setFlag(MboConstants.READONLY, true);
				quantity.setFlag(MboConstants.READONLY, false);
				break;
			case QUANTITY:
				quantity.setFlag(MboConstants.READONLY, true);
				hours.setFlag(MboConstants.READONLY, false);
				break;
			case BOTH:
				hours.setFlag(MboConstants.READONLY, true);
				quantity.setFlag(MboConstants.READONLY, true);
				break;
			}
			
		} 
	}
	
	@Override
	public void action() throws MXException, RemoteException {
		super.action();
		MboValue calcExpression 		= getMboValue(DJPConstants.CALCEXPRESSION);
		MboValue calcExpressionHours 	= getMboValue(DJPConstants.CALCEXPRESSIONHOURS);
		MboValue calcLevel      		= getMboValue(DJPConstants.CALCLEVEL);
		MboValue calcApplyTo			= getMboValue(DJPConstants.CALCAPPLYTO);
		
		Translate t = getTranslator();
		String calcMethod = t.toInternalString("CALCMETHOD", mboValue.getString());
		
		String applyTo = t.toInternalString("CALCAPPLYTO", calcApplyTo.getString());
		boolean isBoth = DJPConstants.CalcApplyTo.BOTH.toString().equals(applyTo);
		
		calcLevel.setValueNull(NOACCESSCHECK);
		calcExpression.setValueNull(NOACCESSCHECK);
		calcExpressionHours.setValueNull(NOACCESSCHECK);
		
		calcExpression.setFlag(REQUIRED, DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE.equals(calcMethod) || DJPConstants.CALCMETHOD_PROPORTIONAL_ATTRIBUTE.equals(calcMethod));
		calcExpression.setFlag(READONLY, DJPConstants.CALCMETHOD_STATIC.equals(calcMethod) || DJPConstants.CALCMETHOD_PROPORTIONAL.equals(calcMethod) || DJPConstants.CALCMETHOD_LEVEL.equals(calcMethod));
		
		calcExpressionHours.setFlag(REQUIRED, isBoth && (DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE.equals(calcMethod) || DJPConstants.CALCMETHOD_PROPORTIONAL_ATTRIBUTE.equals(calcMethod)));
		calcExpressionHours.setFlag(READONLY, !isBoth ||  (DJPConstants.CALCMETHOD_STATIC.equals(calcMethod) || DJPConstants.CALCMETHOD_PROPORTIONAL.equals(calcMethod) || DJPConstants.CALCMETHOD_LEVEL.equals(calcMethod)));
		
		calcLevel.setFlag(REQUIRED, DJPConstants.CALCMETHOD_LEVEL.equals(calcMethod) || DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE.equals(calcMethod));
		calcLevel.setFlag(READONLY, DJPConstants.CALCMETHOD_STATIC.equals(calcMethod) || DJPConstants.CALCMETHOD_PROPORTIONAL.equals(calcMethod) || DJPConstants.CALCMETHOD_PROPORTIONAL_ATTRIBUTE.equals(calcMethod));
		
		if ( !DJPConstants.CALCMETHOD_LEVEL.equalsIgnoreCase(calcMethod) && !DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE.equalsIgnoreCase(calcMethod)) {
			MboValue hours = getMboValue("HOURS");
			MboValue quantity = getMboValue("ITEMQTY");
			hours.setFlag(READONLY, false);
			quantity.setFlag(READONLY, false);
			
		} else {
			MboValue toolHours = getMboValue("HOURS");
			MboValue quantity = getMboValue("ITEMQTY");
			
			String calcApplyToStr = t.toInternalString("CALCAPPLYTO", getMboValue(DJPConstants.CALCAPPLYTO).getString());
			switch(CalcApplyTo.fromStringTo(calcApplyToStr)) {
			case HOURS:
				toolHours.setValue(1.0, MboConstants.NOACCESSCHECK);
				toolHours.setFlag(MboConstants.READONLY, true);
				quantity.setFlag(MboConstants.READONLY, false);
				break;
			case QUANTITY:
				quantity.setValue(1.0, MboConstants.NOACCESSCHECK);
				quantity.setFlag(MboConstants.READONLY, true);
				toolHours.setFlag(MboConstants.READONLY, false);
				break;
			case BOTH:
				toolHours.setValue(1.0, MboConstants.NOACCESSCHECK);
				toolHours.setFlag(MboConstants.READONLY, true);
				quantity.setValue(1.0, MboConstants.NOACCESSCHECK);
				quantity.setFlag(MboConstants.READONLY, true);
				break;
			}
		}
		calcApplyTo.setFlag(READONLY, DJPConstants.CALCMETHOD_STATIC.equals(calcMethod));
	}

}
