package com.ibm.swg.app.jobplan.virtual;

import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.script.ScriptException;

import psdi.app.asset.AssetRemote;
import psdi.app.jobplan.JobPlanRemote;
import psdi.app.jobplan.JobPlanSetRemote;
import psdi.app.workorder.WOSetRemote;
import psdi.mbo.MboConstants;
import psdi.mbo.MboRemote;
import psdi.mbo.MboSet;
import psdi.mbo.MboSetRemote;
import psdi.mbo.NonPersistentMbo;
import psdi.util.MXApplicationException;
import psdi.util.MXException;
import psdi.util.logging.MXLogger;
import psdi.util.logging.MXLoggerFactory;

import com.ibm.swg.app.resourcelevels.ResourceByLevelRemote;
import com.ibm.swg.app.resourcelevels.ResourceLevelsRemote;
import com.ibm.swg.app.resourcelevels.ResourceLevelsSetRemote;
import com.ibm.swg.util.DJPConstants;
import com.ibm.swg.util.DynamicJobPlanHelper;
import com.ibm.swg.util.LinearSegment;
import com.ibm.swg.util.PreparedScript;
import com.ibm.swg.util.PreparedScriptHelper;
import com.ibm.swg.util.DJPConstants.CalcApplyTo;

public class Simulation extends NonPersistentMbo implements SimulationRemote {

	
	public Simulation(MboSet ms) throws RemoteException {
		super(ms);
	}
	
	@Override
	public void applyJobPlan(JobPlanRemote jobplan) throws RemoteException, MXException {
		setValue(JPNUM, jobplan.getString("JPNUM"));
		setValue(PLUSCREVNUM, jobplan.getString("PLUSCREVNUM"));
		setValue(ORGID, jobplan.getString("ORGID"));
		setValue(SITEID, jobplan.getString("SITEID"));
		setValue(ESTDUR, jobplan.getDouble("JPDURATION"));
		setValue(UNITSOFWORK, jobplan.getString(DJPConstants.UNITSOFWORK));
		
		MboSetRemote jpTasks 		= jobplan.getMboSet(JobPlanSetRemote.JOBTASK);
		MboSetRemote jpLabors		= jobplan.getMboSet(JobPlanSetRemote.JOBLABOR);
		MboSetRemote jpMaterials	= jobplan.getMboSet(JobPlanSetRemote.JOBMATERIAL);
		MboSetRemote jpServices		= jobplan.getMboSet(JobPlanSetRemote.JOBSERVICE);
		MboSetRemote jpTools		= jobplan.getMboSet(JobPlanSetRemote.JOBTOOL);
		
		MboSetRemote tasks 		= getMboSet(TASKS);
		MboSetRemote labors 	= getMboSet(LABORS);
		MboSetRemote materials	= getMboSet(MATERIALS);
		MboSetRemote services	= getMboSet(SERVICES);
		MboSetRemote tools		= getMboSet(TOOLS);
		
		for(MboRemote task = jpTasks.moveFirst(); task != null; task = jpTasks.moveNext()) {
			MboRemote newTask = tasks.add();
			newTask.setValue("SIMULATIONNUM", getString("SIMULATIONNUM"));
			newTask.setValue(TYPE, "WOACTIVITY");
			newTask.setValue(JOBRESOURCEID, task.getUniqueIDValue());
		}

		for(MboRemote labor = jpLabors.moveFirst(); labor != null; labor = jpLabors.moveNext()) {
			MboRemote newLabor = labors.add();
			newLabor.setValue("SIMULATIONNUM", getString("SIMULATIONNUM"));
			newLabor.setValue(TYPE, "WPLABOR");
			newLabor.setValue(JOBRESOURCEID, labor.getUniqueIDValue());
		}
		
		for(MboRemote material = jpMaterials.moveFirst(); material != null; material = jpMaterials.moveNext()) {
			MboRemote newMaterial = materials.add();
			newMaterial.setValue("SIMULATIONNUM", getString("SIMULATIONNUM"));
			newMaterial.setValue(TYPE, "WPMATERIAL");
			newMaterial.setValue(JOBRESOURCEID, material.getUniqueIDValue());
		}
		
		for(MboRemote service = jpServices.moveFirst(); service != null; service = jpServices.moveNext()) {
			MboRemote newService = services.add();
			newService.setValue("SIMULATIONNUM", getString("SIMULATIONNUM"));
			newService.setValue(TYPE, "WPSERVICE");
			newService.setValue(JOBRESOURCEID, service.getUniqueIDValue());
		}
		
		for(MboRemote tool = jpTools.moveFirst(); tool != null; tool = jpTools.moveNext()) {
			MboRemote newTool = tools.add();
			newTool.setValue("SIMULATIONNUM", getString("SIMULATIONNUM"));
			newTool.setValue(TYPE, "WPTOOL");
			newTool.setValue(JOBRESOURCEID, tool.getUniqueIDValue());
		}
	}

	private static final Map<String,String[]> quantityFields = new HashMap<String, String[]>();
	private static final Map<String,String[]> hoursFields = new HashMap<String, String[]>();
	private static final Map<String, String> mboReferences = new HashMap<String, String>();
	
		
	MXLogger myLogger = MXLoggerFactory.getLogger("maximo.application.WOTRACK");
	
	/**
	 * Maps for correspondence between JOBPLAN and WORKPLAN
	 */
	static {
		// quantityFields = OBJECTNAME, [RESULT FIELD ON OBJECTNAME, SOURCE FIELD ON ASSOCIATED JOB RESOURCE]
		// if we are calculating the LABOR resource values
		// we will set the value quantityFields.get("WPLABOR")[0]
		// according to the value of the associated Job Resource jobres.getValue(quantityFields.get("WPLABOR")[1])
		quantityFields.put("WPLABOR", new String[]{"QUANTITY","QUANTITY"});
		quantityFields.put("WPTOOL", new String[]{"ITEMQTY","ITEMQTY"});
		quantityFields.put("WPSERVICE", new String[]{"ITEMQTY","ITEMQTY"});
		quantityFields.put("WPMATERIAL", new String[]{"ITEMQTY","ITEMQTY"});
		// quantityFields.put("WORKORDER", new String[]{"",""});		// NO SENSE ! 
		// quantityFields.put("WOACTIVITY", new String[]{"",""});		// NO SENSE !
		// quantityFields.put("WO", new String[]{"",""});				// NO SENSE !
		
		hoursFields.put("WPLABOR", new String[]{"LABORHRS","LABORHRS"});
		hoursFields.put("WPTOOL", new String[]{"HOURS","HOURS"});
		// hoursFields.put("WPSERVICE", new String[]{"",""});			// NO SENSE !
		// hoursFields.put("WPMATERIAL", new String[]{"",""});			// NO SENSE !
		hoursFields.put("WORKORDER", new String[]{"ESTDUR","TASKDURATION"});		 
		hoursFields.put("WOACTIVITY", new String[]{"ESTDUR","TASKDURATION"});
		hoursFields.put("WO", new String[]{"ESTDUR","JPDURATION"});
		
		mboReferences.put("WPLABOR", JobPlanSetRemote.JOBLABOR);
		mboReferences.put("WPTOOL", JobPlanSetRemote.JOBTOOL);
		mboReferences.put("WPSERVICE", JobPlanSetRemote.JOBSERVICE);
		mboReferences.put("WPMATERIAL", JobPlanSetRemote.JOBMATERIAL);
		mboReferences.put("WORKORDER", JobPlanSetRemote.JOBTASK);
		mboReferences.put("WOACTIVITY", JobPlanSetRemote.JOBTASK);
		
	}
	
	
	/**
	 * Main calculation Entry Point
	 * For each resource, we trigger the calculation of the dynamic resources values
	 * Each resource : All Tasks, All labors, All materials, All Tools, All Services, And the workorder itself
	 * For each records of the mbosets, it call the corresponding calculation method, according to the field CALCMETHOD, on
	 * the  Job Plan record (
	 * 	For TASKS 		-> JOBTASK.CALCMETHOD, 
	 * 	For LABOR 		-> JOBLABOR.CALCMETHOD,
	 * 	For MATERIAL 	-> JOBMATERIAL.CALCMETHOD,
	 * 	For TOOL		-> JOBTOOL.CALCMETHOD,
	 * 	FOR SERVICE		-> JOBSERVICE.CALCMETHOD,
	 * 	FOR WORKODER	-> JOBPLAN.CALCMETHOD
	 * )
	 * with parameters : 
	 * 	mbo is the workplan record (wplabor, wpmaterial, wptool, wpservice, workorder)
	 * 	job plan associated record (
	 * 		wplabor 	-> joblabor,
	 * 		wpmaterial 	-> jobmaterial,
	 * 		wptool 		-> jobtool,
	 * 		wpservice 	-> jobservice,
	 * 		workorder	-> jobplan
	 * 	)
	 * calculatedField : the names of the fields on the mbo and job plan associated records
	 * 					 that must be used to set and retrieve the values
	 * 					Example : if JOBRESOURCE.CALCAPPLYTO = HOURS, then we will return for WPLABOR the fields WPLABOR.LABORHRS and JOBLABOR.LABORHRS
	 * 					Example : if JOBRESOURCE.CALCAPPLYTO = HOURS, then we will return for WORKORDER the fields WORKORDER.ESTDUR and JOBPLAN.TASKDURATION
	 * 
	 * @throws MXApplicationException if the workorder is not WAPPR
	 */
	@Override
	public void calculateDynamicJobPlan() throws RemoteException, MXException {
		// Here we can calculate the dynamic job plan values only if the WO is wappr for approval
		
		// Here, for each WPLABOR, WPMATERIAL, WPSERVICE and WPTOOL, we have to calculate the quantity or duration
		// according to the method
		MboSetRemote[] resources = {
				this.getMboSet(TASKS),
				this.getMboSet(LABORS),
				this.getMboSet(MATERIALS),
				this.getMboSet(SERVICES),
				this.getMboSet(TOOLS)
				};
		
		calculateDynamicJobPlan(this, "WO");
		
		for(MboSetRemote mboSet : resources ) {
			//mboSet.reset();
			for(MboRemote mbo = mboSet.moveFirst(); mbo != null; mbo = mboSet.moveNext()) {
				if ( !mbo.isNull(TYPE) ) {
					calculateDynamicJobPlan(mbo, mbo.getString(TYPE));
				}
				
			}
		}
	}
	
	/**
	 * @see#calculateDynamicJobPlan()
	 * @param mbo the resource for which the calculation must be performed
	 * @param objectName the name of the object. Should be obtained by the getMboSetInfo().getObjectName() method, but could be overriden
	 * This name is used to get the fields in mboReferences, hoursFields and quantityFields maps.
	 * @throws RemoteException
	 * @throws MXException
	 */
	private void calculateDynamicJobPlan(MboRemote mbo, String objectName) throws RemoteException, MXException {
		MboRemote jobMboReference = getJobPlanMboReference(mbo, objectName);	// Could be JOBTASK, JOBLABOR, JOBITEM
		if ( jobMboReference == null ) return;
		
		String[] calculatedField = null;//this.getCalculatedFieldName(jobMboReference, objectName);
		boolean isHour = false;
		boolean isBoth = false;
		
		switch( mbo.getClass().equals(getClass()) ? CalcApplyTo.HOURS : DynamicJobPlanHelper.getApplyTo(jobMboReference) ) {
		case HOURS:
			calculatedField = hoursFields.get(objectName);
			isHour = true;
			break;
		case BOTH:
			isBoth = true;
		case QUANTITY:
			calculatedField = quantityFields.get(objectName);
			isHour = false;
			break;
		}
		
		if ( calculatedField == null ) return;	// Maybe we should raise an exception cause the data are not valid

		// This block is executed in priority with the QUANTITY parameter (if the jobplan must apply to BOTH quantity and hours)
		// Then, if must apply to BOTH the second block will be applied only with HOURS param
		String expression;
		if ( jobMboReference.getString(DJPConstants.CALCMETHOD).equals(DJPConstants.CALCMETHOD_PROPORTIONAL)) {
			calculateProportional(mbo, jobMboReference, calculatedField);
		} else if ( jobMboReference.getString(DJPConstants.CALCMETHOD).equals(DJPConstants.CALCMETHOD_LEVEL)) {
			calculateLevel(mbo, jobMboReference, calculatedField, isHour);
		} else if ( jobMboReference.getString(DJPConstants.CALCMETHOD).equals(DJPConstants.CALCMETHOD_PROPORTIONAL_ATTRIBUTE)) {
			expression = jobMboReference.getString(isHour && isBoth? DJPConstants.CALCEXPRESSIONHOURS : DJPConstants.CALCEXPRESSION);
			
			calculateProportionalExpression(mbo, jobMboReference, calculatedField, expression);
		} else if ( jobMboReference.getString(DJPConstants.CALCMETHOD).equals(DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE)) {
			expression = jobMboReference.getString(isHour && isBoth ? DJPConstants.CALCEXPRESSIONHOURS : DJPConstants.CALCEXPRESSION) ;
			calculateLevelAttribute(mbo, jobMboReference, calculatedField, expression, isHour);
		}
		
		if ( isBoth && mbo.getClass().equals(getClass()) || DynamicJobPlanHelper.getApplyTo(jobMboReference) == CalcApplyTo.BOTH ) {
			calculatedField = hoursFields.get(objectName);
			if ( calculatedField == null ) return;	// Maybe we should raise an exception cause the data are not valid
			
			if ( jobMboReference.getString(DJPConstants.CALCMETHOD).equals(DJPConstants.CALCMETHOD_PROPORTIONAL)) {
				calculateProportional(mbo, jobMboReference, calculatedField);
			} else if ( jobMboReference.getString(DJPConstants.CALCMETHOD).equals(DJPConstants.CALCMETHOD_LEVEL)) {
				calculateLevel(mbo, jobMboReference, calculatedField, true);
			} else if ( jobMboReference.getString(DJPConstants.CALCMETHOD).equals(DJPConstants.CALCMETHOD_PROPORTIONAL_ATTRIBUTE)) {
				expression = jobMboReference.getString(DJPConstants.CALCEXPRESSIONHOURS);
				calculateProportionalExpression(mbo, jobMboReference, calculatedField, expression);
			} else if ( jobMboReference.getString(DJPConstants.CALCMETHOD).equals(DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE)) {
				expression = jobMboReference.getString(DJPConstants.CALCEXPRESSIONHOURS);
				calculateLevelAttribute(mbo, jobMboReference, calculatedField, expression, true);
			}
		}
	}
	
	/**
	 * Method use to detect if the job plan is set as dynamic and should be apply using the length of the 
	 * workorder (startmeasure and endmeasure) or not.
	 * If not, the TOTAL WORK UNITS of the jobresources will be use to calculate the results.
	 * @return true if the jobplan is dynamic and should apply the linear calculation. Otherwise return false
	 * @throws RemoteException
	 * @throws MXException
	 */
	public boolean mustApplyLinearCalculation() throws RemoteException, MXException {
		MboRemote jobplan = getJobPlanMboReference(null, "WO");
		return !jobplan.getBoolean(DJPConstants.APPLYNONDYNAMICCALC);
	}
	
	private double calculateLevelAttributeForMultiAsset(MboRemote multiAsset, MboRemote plan, MboRemote jobres, String[] calculatedField, String expression, boolean isHour) throws RemoteException, MXException {
		if ( multiAsset.toBeDeleted() ) return 0;
		AssetRemote asset = (AssetRemote) multiAsset.getMboSet("ASSET").getMbo(0);
		if ( asset == null ) return 0;
		
		if ( asset.getBoolean("ISLINEAR") && mustApplyLinearCalculation()) {
			return calculateProportionalExpressionOnLinearAsset(multiAsset, asset, jobres, calculatedField[1], expression);
		} else {
			double totalWorkUnits = multiAsset.getDouble(TOTALWORKUNITS);
			return calculateProportionalExpressionOnNonLinearAsset(asset, jobres, calculatedField[1], expression, totalWorkUnits);
		}
	}
	
	/**
	 * 
	 * @param plan	
	 * 		the workplan record (could be WPLABOR, WPMATERIAL, WPTOOL, WPSERVICE, WORKORDER)
	 * @param 
	 * 		jobres the job plan associated records (could be JOBLABOR, JOBMATERIAL, JOBTOOL, JOBSERVICE, JOBPLAN)
	 * @param 
	 * 		calculatedField the names of the fields on the mbo and job plan associated records
	 * 
	 * @see DJPWO#calculateDynamicJobPlan()
	 * 
	 * @throws RemoteException
	 * @throws MXException
	 * 		for now it throw a NotYetImplemented Exception
	 */
	private void calculateLevelAttribute(MboRemote plan, MboRemote jobres, String[] calculatedField, String expression, boolean isHour) throws RemoteException, MXException {
		// for each linear asset
		MboSetRemote multiAssetLocCISet = getMboSet("ALLMULTIASSETLOCCI");
//		multiAssetLocCISet.reset();
//		multiAssetLocCISet.setWhere("assetnum is not null");
		
		ResourceLevelsSetRemote resourceLevels = (ResourceLevelsSetRemote) jobres.getMboSet("RESLEVEL");
		resourceLevels.reset();
		ResourceLevelsRemote resourceLevel = (ResourceLevelsRemote) resourceLevels.getMbo(0);
		String unitOfResourceLevels = resourceLevel.getString(ResourceLevelsRemote.MEASUREUNITID);
		
		double totalResourceQty = 0;
		
		for(MboRemote multiAsset = multiAssetLocCISet.moveFirst(); multiAsset != null; multiAsset = multiAssetLocCISet.moveNext()) {
			if ( multiAsset.toBeDeleted() ) continue;
			double tmp = calculateLevelAttributeForMultiAsset(multiAsset, plan, jobres, calculatedField, expression, isHour);
			tmp = DynamicJobPlanHelper.convertValue(this, tmp, getString(DJPConstants.UNITSOFWORK), unitOfResourceLevels);
			ResourceByLevelRemote resourceByLevel = resourceLevel.getLevelForValue(tmp);
			
			if (resourceByLevel == null) {
				throw new MXApplicationException("dynamicjobplan","invalidResourceLevel", new Object[]{totalResourceQty,resourceLevel.getResourceLevelsNum()});
			}
			totalResourceQty += isHour ? resourceByLevel.getHours() : resourceByLevel.getQuantity();
		}
		multiAssetLocCISet.close();

		plan.setValue(calculatedField[0], totalResourceQty);
//		if( myLogger.isInfoEnabled() ) {
//			myLogger.info("Calculate Level value for " + plan.getThisMboSet().getMboSetInfo().getObjectName() + " " + plan.getUniqueIDValue() + " with Level " + resourceLevels.getString(ResourceLevelsRemote.RESOURCELEVELSNUM));
//		}
//		
//		if ( unitOfResourceLevels != null && !unitOfResourceLevels.isEmpty() ) {
//			totalResourceQty = DynamicJobPlanHelper.convertValue(this, totalResourceQty, getString(DJPConstants.UNITSOFWORK), unitOfResourceLevels);
//		}
//		
//		if( myLogger.isInfoEnabled() ) {
//			myLogger.info("\tvalue = " + totalResourceQty);
//		}
//		
//		ResourceByLevelRemote resourceByLevel = (resourceLevels).getLevelForValue(totalResourceQty);
//		
//		if (resourceByLevel == null) {
//			throw new MXApplicationException("dynamicjobplan","invalidResourceLevel", new Object[]{totalResourceQty,resourceLevels.getResourceLevelsNum()});
//		}
//		
//		if ( isHour ) {
//			plan.setValue(calculatedField[0], resourceByLevel.getHours());
//		} else {
//			plan.setValue(calculatedField[0], resourceByLevel.getQuantity());
//		}
	}

	
	/**
	 *  According to the MAXVAR "DYNAMIC_ERROR", which could be CONTINUE or STOP
	 * @return true is the maxvar value is STOP, false otherwise
	 * @throws RemoteException
	 * @throws MXException
	 */
	private boolean stopOnError() throws RemoteException, MXException {
		return "STOP".equalsIgnoreCase(getMboServer().getMaxVar().getString("DYNAMIC_ONERROR", getOrgSiteForMaxvar("DYNAMIC_ONERROR")));
	}

	
	private double calculateProportionalExpressionOnLinearAsset(MboRemote multiAsset, AssetRemote asset, MboRemote jobres, String calculatedField, String expression) throws RemoteException, MXException {
		String unitsOfWork = getString(DJPConstants.UNITSOFWORK);
		double jobplanUnitOfWork = DynamicJobPlanHelper.getJobPlan(jobres).getDouble(DJPConstants.UNITSOFWORKMULTIPLIER);
		jobplanUnitOfWork = DynamicJobPlanHelper.convertValue(this, jobplanUnitOfWork, DynamicJobPlanHelper.getJobPlan(jobres).getString(DJPConstants.UNITSOFWORK), unitsOfWork);
		
		// Get the workorder start measure and end measure
		double woStartMeasure = multiAsset.getDouble("STARTMEASURE");
		double woEndMeasure = multiAsset.getDouble("ENDMEASURE");
		String startMeasureUnitId = multiAsset.getString("STARTMEASUREUNITID");
		String endMeasureUnitId = multiAsset.getString("ENDMEASUREUNITID");
		
		woStartMeasure = DynamicJobPlanHelper.convertValue(this, woStartMeasure, startMeasureUnitId, unitsOfWork);
		woEndMeasure = DynamicJobPlanHelper.convertValue(this, woEndMeasure, endMeasureUnitId, unitsOfWork);
		
		float resourceQty = 0;
		Double res = null;
//		String expression = jobres.getString(DJPConstants.CALCEXPRESSION);
		PreparedScript<Double> script = new PreparedScript<Double>(expression);
		List<LinearSegment> segments = PreparedScriptHelper.getSegments(script, asset, woStartMeasure, woEndMeasure, unitsOfWork);
		
		for(LinearSegment segment : segments) {
			if ( segment.getStart() >= woEndMeasure || segment.getEnd() <= woStartMeasure ) continue;
			script.clearParameterValues();
			
//			script.setParameterValue(DJPConstants.HOURS_TAG, jobres.getDouble(calculatedField));
//			script.setParameterValue(DJPConstants.QUANTITY_TAG, jobres.getDouble(calculatedField));

			for(String param : script.getParametersName()) {
//				if ( param.equals(DJPConstants.HOURS_TAG) || param.equals(DJPConstants.QUANTITY_TAG) ) continue;
				if ( segment.attributeNames().contains(param) ) {
					script.setParameterValue(param, segment.getAttribute(param));
				} else {
					if ( stopOnError() )
						throw new MXApplicationException("dynamicjobplan","attributeMissing",new Object[]{param, segment.getStart(), segment.getEnd()});
				}
			}
			
			try {
				double tmpStart = Math.max(segment.getStart(),woStartMeasure);
				double tmpEnd = Math.min(segment.getEnd(), woEndMeasure);
				//double distance = Math.max(segment.getEnd(),segment.getStart()) - Math.min(segment.getEnd(),  segment.getStart());
				double distance = Math.max(tmpStart, tmpEnd) - Math.min(tmpStart, tmpEnd);
				double jobplanQty = jobres.getDouble(calculatedField);
				res = jobplanQty * (Double)script.eval() * (distance) / jobplanUnitOfWork;
				resourceQty += res;
			} catch (ScriptException e) {
				if ( stopOnError() )
					throw new MXApplicationException("dynamicjobplan","invalidExpression",e);
			}
			
		}
		
		return resourceQty;
	}
	
	
	private double calculateProportionalExpressionOnNonLinearAsset(AssetRemote asset, MboRemote jobres, String calculatedField, String expression, double totalWorkUnits) throws RemoteException, MXException {
		String unitsOfWork = getString(DJPConstants.UNITSOFWORK);
//		double totalWorkUnits = getDouble(TOTALWORKUNITS);
		double jobplanUnitOfWork = DynamicJobPlanHelper.getJobPlan(jobres).getDouble(DJPConstants.UNITSOFWORKMULTIPLIER);
		jobplanUnitOfWork = DynamicJobPlanHelper.convertValue(this, jobplanUnitOfWork, DynamicJobPlanHelper.getJobPlan(jobres).getString(DJPConstants.UNITSOFWORK), unitsOfWork);
		
//		String expression = jobres.getString(DJPConstants.CALCEXPRESSION);
		Double res = null;
		
		// Here we should not set the value for both, but the Adapter on the JOBRESOURCE.CALCEXPRESSION should catch a wrong expression
		// i.e an expression that should not use one of HOURS or QUANTITY TAG
		PreparedScript<Double> script = new PreparedScript<Double>(expression);
		LinearSegment segment = PreparedScriptHelper.getSegment(script, asset);
		
//		script.setParameterValue(DJPConstants.HOURS_TAG, jobres.getDouble(calculatedField));
//		script.setParameterValue(DJPConstants.QUANTITY_TAG, jobres.getDouble(calculatedField));

		for(String param : script.getParametersName()) {
//			if ( param.equals(DJPConstants.HOURS_TAG) || param.equals(DJPConstants.QUANTITY_TAG) ) continue;
			if ( segment != null && segment.getAttribute(param) != null ) {
				script.setParameterValue(param, segment.getAttribute(param));
			} else {
				if ( stopOnError() )
					throw new MXApplicationException("dynamicjobplan","attributeMissing",new Object[]{param, segment.getStart(), segment.getEnd()});
			}
		}
		
		try {
			res = (Double)script.eval();
			res = res * jobres.getDouble(calculatedField);
			res = res * totalWorkUnits / jobplanUnitOfWork;
		} catch (ScriptException e) {
			if ( stopOnError())
				throw new MXApplicationException("dynamicjobplan","invalidExpression",e);
		}
		
		return res;
	}
	
	/**
	 * For each asset/location/CI apply the dynamic jobplan.
	 * @param plan
	 * @param jobres
	 * @param calculatedField
	 * @param expression
	 * @throws RemoteException
	 * @throws MXException
	 */
	private void calculateProportionalExpression(MboRemote plan, MboRemote jobres, String[] calculatedField, String expression) throws RemoteException, MXException {		
		// for each linear asset
		MboSetRemote multiAssetLocCISet = getMboSet("ALLMULTIASSETLOCCI");
//		multiAssetLocCISet.reset();
//		multiAssetLocCISet.setWhere("assetnum is not null");
		
		double totalResourceQty = 0;
		
		for(MboRemote multiAsset = multiAssetLocCISet.moveFirst(); multiAsset != null; multiAsset = multiAssetLocCISet.moveNext()) {
			if ( multiAsset.toBeDeleted() ) continue;
			AssetRemote asset = (AssetRemote) multiAsset.getMboSet("ASSET").getMbo(0);

			double resourceQty = 0;
			if ( asset.getBoolean("ISLINEAR") && mustApplyLinearCalculation() ) {
				resourceQty = calculateProportionalExpressionOnLinearAsset(multiAsset, asset, jobres, calculatedField[1], expression);
			} else {
				double totalWorkUnits = multiAsset.getDouble(TOTALWORKUNITS);
				resourceQty = calculateProportionalExpressionOnNonLinearAsset(asset, jobres, calculatedField[1], expression, totalWorkUnits);
			}
			totalResourceQty += resourceQty;
		}
//		multiAssetLocCISet.close();
		
		plan.setValue(calculatedField[0], totalResourceQty);
	}
	
	private void calculateProportional(MboRemote plan, MboRemote jobres, String[] calculatedField) throws RemoteException, MXException {
		String unitsOfWork = getString(DJPConstants.UNITSOFWORK);
		double jobplanQty = jobres.getDouble(calculatedField[1]);
		double jobplanUnitOfWork = DynamicJobPlanHelper.getJobPlan(jobres).getDouble(DJPConstants.UNITSOFWORKMULTIPLIER);
		jobplanUnitOfWork = DynamicJobPlanHelper.convertValue(this, jobplanUnitOfWork, DynamicJobPlanHelper.getJobPlan(jobres).getString(DJPConstants.UNITSOFWORK), getString(DJPConstants.UNITSOFWORK));
		
		if( myLogger.isInfoEnabled() ) {
			myLogger.info("Calculate proportional value for " + plan.getThisMboSet().getMboSetInfo().getObjectName() + " " + plan.getUniqueIDValue() );
		}
		
		// for each linear asset
		MboSetRemote multiAssetLocCISet = getMboSet("ALLMULTIASSETLOCCI");
//		multiAssetLocCISet.reset();
				
		double totalResourceQty = 0;
//		boolean linearAsset = false;
//		boolean nonLinearAsset = false;
		for(MboRemote multiAsset = multiAssetLocCISet.moveFirst(); multiAsset != null; multiAsset = multiAssetLocCISet.moveNext()) {
			if ( multiAsset.toBeDeleted() ) continue;
			AssetRemote asset = (AssetRemote) multiAsset.getMboSet("ASSET").getMbo(0);

			double resourceQty = 0;
			if ( asset != null && asset.getBoolean("ISLINEAR") && mustApplyLinearCalculation() ) {
//				linearAsset = true;
//				if ( nonLinearAsset ) {
//					throw new MXApplicationException("dynamicjobplan","linearAndNonLinear");
//				}
				double woStartMeasure = multiAsset.getDouble("STARTMEASURE");
				double woEndMeasure = multiAsset.getDouble("ENDMEASURE");
				String startMeasureUnitId = multiAsset.getString("STARTMEASUREUNITID");
				String endMeasureUnitId = multiAsset.getString("ENDMEASUREUNITID");
				
				woStartMeasure = DynamicJobPlanHelper.convertValue(this, woStartMeasure, startMeasureUnitId, unitsOfWork);
				woEndMeasure = DynamicJobPlanHelper.convertValue(this, woEndMeasure, endMeasureUnitId, unitsOfWork);
				
				resourceQty = jobplanQty * Math.abs(woStartMeasure - woEndMeasure) / jobplanUnitOfWork;
			} else {
//				nonLinearAsset = true;
//				if ( linearAsset ) {
//					throw new MXApplicationException("dynamicjobplan","linearAndNonLinear");
//				}
				
//				double totalWorkUnits = getDouble(TOTALWORKUNITS);
				double totalWorkUnits = multiAsset.getDouble(TOTALWORKUNITS);
				resourceQty = jobplanQty * totalWorkUnits / jobplanUnitOfWork;
			}
			totalResourceQty += resourceQty;
		}
//		multiAssetLocCISet.close();
		
		plan.setValue(calculatedField[0], totalResourceQty);
		
		
	}
	
	private void calculateLevel(MboRemote plan, MboRemote jobres, String[] calculatedField, boolean isHour) throws RemoteException, MXException {
		
		ResourceLevelsSetRemote resourceLevels = (ResourceLevelsSetRemote) jobres.getMboSet("RESLEVEL");
		resourceLevels.reset();
		ResourceLevelsRemote resourceLevel = (ResourceLevelsRemote) resourceLevels.getMbo(0);
		String unitOfResourceLevels = resourceLevel.getString(ResourceLevelsRemote.MEASUREUNITID);
		
		MboSetRemote multiAssetLocCISet = getMboSet("ALLMULTIASSETLOCCI");
//		multiAssetLocCISet.reset();
		
		double resourceQty = 0;
		for(MboRemote multiAsset = multiAssetLocCISet.moveFirst(); multiAsset != null; multiAsset = multiAssetLocCISet.moveNext()) {
			if ( multiAsset.toBeDeleted() ) continue;
			double tmp = DynamicJobPlanHelper.convertValue(this, multiAsset.getDouble(TOTALWORKUNITS), getString(DJPConstants.UNITSOFWORK), unitOfResourceLevels);
			ResourceByLevelRemote resourceByLevel = resourceLevel.getLevelForValue(tmp);
			
			if (resourceByLevel == null) {
				throw new MXApplicationException("dynamicjobplan","invalidResourceLevel");
			}
			
			resourceQty += isHour ? resourceByLevel.getHours() : resourceByLevel.getQuantity();
		}
		plan.setValue(calculatedField[0], resourceQty);
		
	}

	/**
	 * According to the type of object (Aka OBJECTNAME) and the calculation method (Aka CALCMETHOD)
	 * The field that will be calculated will be a quantity field, or a duration field
	 * For Example : <br/>
	 * <ul>
	 * <li>for [JOBLABOR, QUANTITY method] -> Quantity</li>
	 * <li>for [JOBLABOR, HOURS method] -> LABORHRS</li>
	 * <li>for [JOBTOOL, QUANTITY method] -> ItemQty</li>
	 * <li>for [JOBTOOL, HOURS method] -> Hours</li>
	 * <li>...</li>
	 * </ul> 
	 * @param mbo the record that will be calculated. It could be a task/children (WOACTIVITY), a labor (WPLABOR), an item (WPMATERIAL,WPSERVICE, WPTOOL)
	 * @return a couple of value, representing the WORKPLAN destination field name, and JOBPLAN field name that is used as reference
	 * @throws RemoteException
	 * @throws MXException
	 */
//	private String[] getCalculatedFieldName(MboRemote mbo, String objectName) throws RemoteException, MXException {
//		// get the field name that will be calculated, according to the CALCAPPLYTO
//		String[] calculatedField = null;
//		switch(DynamicJobPlanHelper.getApplyTo(mbo)) {
//		case HOURS:
//			calculatedField = hoursFields.get(objectName);
//			break;
//		case QUANTITY:
//			calculatedField = quantityFields.get(objectName);
//			break;
//		case BOTH:
//			String[] hours = hoursFields.get(objectName);
//			String[] qty   = quantityFields.get(objectName);
//			
//			calculatedField = new String[hours.length + qty.length];
//			System.arraycopy(hours, 0, calculatedField, 0, hours.length);
//			System.arraycopy(qty, 0, calculatedField, hours.length - 1, qty.length);
//			
//		default:
//			break;
//		
//		}
//		
//		return calculatedField;
//	}
	
	/**
	 * This method returns the corresponding jobplan record for resources/tasks in the workplan <br/>
	 * Ex : if mbo instanceof WPLABOR -> it returns the JOBLABOR record corresonding to mbo
	 * @param mbo a workplan record (WOACTIVITY/WORKORDER, WPLABOR, WPITEM, WPMATERIAL, WPSERVICE, WPTOOL)
	 * @return the corresponding record in the jobplan
	 * @throws RemoteException
	 * @throws MXException
	 */
	private MboRemote getJobPlanMboReference(MboRemote mbo, String objectName) throws RemoteException, MXException {
		JobPlanRemote jobPlan = (JobPlanRemote) getMboSet(WOSetRemote.JOBPLAN).getMbo(0);
		
		if ( jobPlan == null ) return null;
		if ( objectName.equalsIgnoreCase("WO") ) return jobPlan;
		
		MboSetRemote set = jobPlan.getMboSet(mboReferences.get(objectName));
		return set.getMboForUniqueId(mbo.getLong("JOBRESOURCEID"));

	}
	
	@Override
	public void calculateTotalWorkUnitsFor(MboRemote multiAssetLocCI, String unitOfWork) throws RemoteException, MXException {
		Double endMeasure = DynamicJobPlanHelper.convertValue(this, multiAssetLocCI.getDouble("ENDMEASURE"),multiAssetLocCI.getString("ENDMEASUREUNITID"),unitOfWork);
		Double startMeasure = DynamicJobPlanHelper.convertValue(this, multiAssetLocCI.getDouble("STARTMEASURE"), multiAssetLocCI.getString("STARTMEASUREUNITID"),unitOfWork);
		multiAssetLocCI.setValue(DJPConstants.TOTALWORKUNITS, Math.abs( endMeasure - startMeasure ), MboConstants.NOACCESSCHECK);
	}
	
	
	/**
	 * Calculate the total length of attributes : for each linear asset in the multiassetlocci set,
	 * we add the absolute value for endmeasure - startmeasure (cause the endmeasure is not necessarely 
	 * lower than the startmeasure)<br/>
	 * Update the field TOTALWORKUNITS<br/>
	 * If some assets are not linear, they are skipped, and a warning is added to the ALLMULTIASSETLOCCI set
	 */
	@Override
	public void calculateTotalWorkUnits() throws RemoteException, MXException {
		if ( myLogger.isInfoEnabled() ) {
			myLogger.info("Recalculating Total Work Units based on ALLMULTIASSETLOCCI relationship on assets (assetnum not null)");
		}
		
		if ( !mustApplyLinearCalculation() ) {
			if ( myLogger.isInfoEnabled() ) {
				myLogger.info("Ignoring recalulation of total work units because the jobplan must not be applied as linear");
			}
			getThisMboSet().addWarning(new MXApplicationException("dynamic","notApplicable"));
			return;
		}
		
		MboSetRemote multiAssetLocCISet = getMboSet("ALLMULTIASSETLOCCI");
//		multiAssetLocCISet.resetQbe();
//		double totalWorkUnits = 0;
		
//		// First we take the linear segment details of the workorder
//		MboRemote linear = getMboSet("LINEARRELATED").getMbo(0);
//		if ( isNull(DJPConstants.UNITSOFWORK) ) {
//			//if ( multiAssetLocCISet.count() != 0 ) {
//				// Here : get the unit of measure id from the MULTIASSETLOCCI if linear
//				setValue(DJPConstants.UNITSOFWORK,linear.getString("STARTMEASUREUNITID"));
//			//}
//		}
		
////		Double endMeasure = 0.0;
////		Double startMeasure = 0.0;
//		
//		if ( linear != null ) {
//			calculateTotalWorkUnitsFor(linear, unitOfWork);
//		}

		 
		if ( multiAssetLocCISet.count() > 0 ) {
			if ( isNull(DJPConstants.UNITSOFWORK) ) {
				// Here : get the unit of measure id from the MULTIASSETLOCCI if linear
				setValue(DJPConstants.UNITSOFWORK,multiAssetLocCISet.getMbo(0).getString("STARTMEASUREUNITID"));
			}
			
			String unitOfWork = getString(DJPConstants.UNITSOFWORK);
			for(MboRemote multi = multiAssetLocCISet.moveFirst(); multi != null; multi = multiAssetLocCISet.moveNext()) {
				if ( multi.toBeDeleted()) continue;
				if ( multi.getMboSet("ASSET").getMbo(0) != null && multi.getMboSet("ASSET").getMbo(0).getBoolean("ISLINEAR") ) {
					calculateTotalWorkUnitsFor(multi, unitOfWork);
				}
//				endMeasure = DynamicJobPlanHelper.convertValue(this, multi.getDouble("ENDMEASURE"),multi.getString("ENDMEASUREUNITID"),unitOfWork);
//				startMeasure = DynamicJobPlanHelper.convertValue(this, multi.getDouble("STARTMEASURE"), multi.getString("STARTMEASUREUNITID"),unitOfWork);
//				
//				multi.setValue(DJPConstants.TOTALWORKUNITS, Math.abs( endMeasure - startMeasure ), MboConstants.NOACCESSCHECK);
			}
		}
//		setValue(TOTALWORKUNITS, totalWorkUnits, MboConstants.NOACCESSCHECK);
	}

}
