package com.ibm.swg.app.resourcelevels;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.List;

import psdi.mbo.Mbo;
import psdi.mbo.MboServerInterface;
import psdi.mbo.MboSet;
import psdi.util.MXApplicationException;
import psdi.util.MXException;

public class ResourceByLevelSet extends MboSet implements ResourceByLevelSetRemote {

	public ResourceByLevelSet(MboServerInterface ms) throws RemoteException {
		super(ms);
	}

	@Override
	protected Mbo getMboInstance(MboSet ms) throws MXException, RemoteException {
		return new ResourceByLevel(ms);
	}

	@Override
	public 	ResourceByLevelRemote getLevelForValue(double value) throws RemoteException, MXException {
		reset();
		resetQbe();

		setWhere(value + " > " + ResourceByLevelRemote.LOWERTHRESHOLD + " AND (" + ResourceByLevelRemote.UPPERTHRESHOLD + " is null OR " + value + " <= "+ ResourceByLevelRemote.UPPERTHRESHOLD + ")");
		if ( count() == 0 ) {
			reset();
			resetQbe();
			setWhere(value + " = " + ResourceByLevelRemote.LOWERTHRESHOLD);
		}
		if ( count() == 0 ) {
			reset();
			resetQbe();
			setWhere(ResourceByLevelRemote.LOWERTHRESHOLD + " is null AND " + ResourceByLevelRemote.UPPERTHRESHOLD + " > " + value);
		}
		return (ResourceByLevelRemote) getMbo(0);
		
	}
	
	private static class Segment {
		public double start;
		public double end;
		
		public Segment(double start, double end) {
			this.start = start;
			this.end = end;
		}
		
		public boolean intersect(Segment s) {
			return !(this.start <= s.start && this.end <= s.start && this.end < s.end ||
					s.start <= this.start && s.end <= this.start && s.end < this.end ||
					this.start >= s.end || this.end <= s.start)
					;
			
			//return this.start > s.start && this.start < s.end || this.end > s.start && this.end < s.end;
		}
	}
	
	@Override
	public void validate() throws MXException, RemoteException {
		super.validate();
		List<Segment> segments = new ArrayList<Segment>();
		for(ResourceByLevelRemote mbo = (ResourceByLevelRemote) moveFirst(); mbo != null; mbo = (ResourceByLevelRemote) moveNext()) {
			if ( mbo.toBeDeleted() ) continue;
			
			if ( mbo.isNull(ResourceByLevelRemote.LOWERTHRESHOLD) && mbo.isNull(ResourceByLevelRemote.UPPERTHRESHOLD)) {
				throw new MXApplicationException("dynamicjobplan","ThresholdsInvalid");
			}
			double lower = mbo.isNull(ResourceByLevelRemote.LOWERTHRESHOLD) ? Double.MIN_VALUE : mbo.getLowerThreshold();
			double upper = mbo.isNull(ResourceByLevelRemote.UPPERTHRESHOLD) ? Double.MAX_VALUE : mbo.getUpperThreshold();
			segments.add(new Segment(lower, upper));
		}
		
		Segment head = segments.isEmpty() ? null : segments.remove(0);
		while ( head != null) {
			for(Segment tmp : segments) {
				if ( head.intersect(tmp)) {
					throw new MXApplicationException("dynamicjobplan","ThresholdsInvalid");
				}
			}
			
			if ( !segments.isEmpty() ) {
				head = segments.remove(0);
			} else {
				head = null;
			}
		};
	}
	
	
	public static void main(String[] args) {
		Segment s1, s2;
		
		// cas 1
		s1 = new Segment(1,4);
		s2 = new Segment(1,2);
		System.out.println("Cas 1");
		System.out.println(s1.intersect(s2));
		System.out.println(s2.intersect(s1));
		
		// cas 2
		s1 = new Segment(1,3);
		s2 = new Segment(1,6);
		System.out.println("Cas 2");
		System.out.println(s1.intersect(s2));
		System.out.println(s2.intersect(s1));
				
		// cas 3
		s1 = new Segment(1,4);
		s2 = new Segment(1,4);
		System.out.println("Cas 3");
		System.out.println(s1.intersect(s2));
		System.out.println(s2.intersect(s1));
		
		// cas 4
		s1 = new Segment(1,4);
		s2 = new Segment(2,3);
		System.out.println("Cas 4");
		System.out.println(s1.intersect(s2));
		System.out.println(s2.intersect(s1));
		
		// cas 5
		s1 = new Segment(4,5);
		s2 = new Segment(1,10);
		System.out.println("Cas 5");
		System.out.println(s1.intersect(s2));
		System.out.println(s2.intersect(s1));
				
		// cas 6
		s1 = new Segment(1,4);
		s2 = new Segment(4,8);
		System.out.println("Cas 6");
		System.out.println(s1.intersect(s2));
		System.out.println(s2.intersect(s1));
		
		// cas 7
		s1 = new Segment(1,4);
		s2 = new Segment(2,4);
		System.out.println("Cas 7");
		System.out.println(s1.intersect(s2));
		System.out.println(s2.intersect(s1));
				
		// cas 8
		s1 = new Segment(3,4);
		s2 = new Segment(1,4);
		System.out.println("Cas 8");
		System.out.println(s1.intersect(s2));
		System.out.println(s2.intersect(s1));
				
				
		// cas 9
		s1 = new Segment(1,4);
		s2 = new Segment(5,6);
		System.out.println("Cas 9");
		System.out.println(s1.intersect(s2));
		System.out.println(s2.intersect(s1));
		
		
	}
	
}
