package com.ibm.swg.app.workorder;

import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.script.ScriptException;

import psdi.app.asset.AssetRemote;
import psdi.app.jobplan.JobPlanRemote;
import psdi.app.jobplan.JobPlanSetRemote;
import psdi.app.workorder.WO;
import psdi.app.workorder.WOSetRemote;
import psdi.mbo.MboConstants;
import psdi.mbo.MboRemote;
import psdi.mbo.MboSet;
import psdi.mbo.MboSetRemote;
import psdi.security.UserInfo;
import psdi.server.MXServer;
import psdi.util.MXApplicationException;
import psdi.util.MXApplicationYesNoCancelException;
import psdi.util.MXException;
import psdi.util.logging.MXLogger;
import psdi.util.logging.MXLoggerFactory;

import com.ibm.swg.app.resourcelevels.ResourceByLevelRemote;
import com.ibm.swg.app.resourcelevels.ResourceLevelsRemote;
import com.ibm.swg.app.resourcelevels.ResourceLevelsSetRemote;
import com.ibm.swg.util.DJPConstants;
import com.ibm.swg.util.DynamicJobPlanHelper;
import com.ibm.swg.util.LinearSegment;
import com.ibm.swg.util.PreparedScript;
import com.ibm.swg.util.PreparedScriptHelper;
import com.ibm.swg.util.DJPConstants.CalcApplyTo;

public class DJPWO extends WO implements DJPWORemote {

	private static final Map<String,String[]> quantityFields = new HashMap<String, String[]>();
	private static final Map<String,String[]> hoursFields = new HashMap<String, String[]>();
	private static final Map<String, String> mboReferences = new HashMap<String, String>();
	
	
	MXLogger myLogger = MXLoggerFactory.getLogger("maximo.application.WOTRACK");
	
	/**
	 * Maps for correspondence between JOBPLAN and WORKPLAN
	 */
	static {
		// quantityFields = OBJECTNAME, [RESULT FIELD ON OBJECTNAME, SOURCE FIELD ON ASSOCIATED JOB RESOURCE]
		// if we are calculating the LABOR resource values
		// we will set the value quantityFields.get("WPLABOR")[0]
		// according to the value of the associated Job Resource jobres.getValue(quantityFields.get("WPLABOR")[1])
		quantityFields.put("WPLABOR", new String[]{"QUANTITY","QUANTITY"});
		quantityFields.put("WPTOOL", new String[]{"ITEMQTY","ITEMQTY"});
		quantityFields.put("WPSERVICE", new String[]{"ITEMQTY","ITEMQTY"});
		quantityFields.put("WPMATERIAL", new String[]{"ITEMQTY","ITEMQTY"});
		// quantityFields.put("WORKORDER", new String[]{"",""});		// NO SENSE ! 
		// quantityFields.put("WOACTIVITY", new String[]{"",""});		// NO SENSE !
		// quantityFields.put("WO", new String[]{"",""});				// NO SENSE !
		
		hoursFields.put("WPLABOR", new String[]{"LABORHRS","LABORHRS"});
		hoursFields.put("WPTOOL", new String[]{"HOURS","HOURS"});
		// hoursFields.put("WPSERVICE", new String[]{"",""});			// NO SENSE !
		// hoursFields.put("WPMATERIAL", new String[]{"",""});			// NO SENSE !
		hoursFields.put("WORKORDER", new String[]{"ESTDUR","TASKDURATION"});		 
		hoursFields.put("WOACTIVITY", new String[]{"ESTDUR","TASKDURATION"});
		hoursFields.put("WO", new String[]{"ESTDUR","JPDURATION"});
		
		mboReferences.put("WPLABOR", JobPlanSetRemote.JOBLABOR);
		mboReferences.put("WPTOOL", JobPlanSetRemote.JOBTOOL);
		mboReferences.put("WPSERVICE", JobPlanSetRemote.JOBSERVICE);
		mboReferences.put("WPMATERIAL", JobPlanSetRemote.JOBMATERIAL);
		mboReferences.put("WORKORDER", JobPlanSetRemote.JOBTASK);
		mboReferences.put("WOACTIVITY", JobPlanSetRemote.JOBTASK);
		
	}
	
	private volatile boolean mustApplyDJP = false;
	
	
	public DJPWO(MboSet ms) throws MXException, RemoteException {
		super(ms);
	}

	public void setMustApplyDJP(boolean mustApplyDJP) {
		this.mustApplyDJP = mustApplyDJP;
	}
	
	public boolean mustApplyDJP() {
		return mustApplyDJP;
	}
	
	/**
	 * Main calculation Entry Point
	 * For each resource, we trigger the calculation of the dynamic resources values
	 * Each resource : All Tasks, All labors, All materials, All Tools, All Services, And the workorder itself
	 * For each records of the mbosets, it call the corresponding calculation method, according to the field CALCMETHOD, on
	 * the  Job Plan record (
	 * 	For TASKS 		-> JOBTASK.CALCMETHOD, 
	 * 	For LABOR 		-> JOBLABOR.CALCMETHOD,
	 * 	For MATERIAL 	-> JOBMATERIAL.CALCMETHOD,
	 * 	For TOOL		-> JOBTOOL.CALCMETHOD,
	 * 	FOR SERVICE		-> JOBSERVICE.CALCMETHOD,
	 * 	FOR WORKODER	-> JOBPLAN.CALCMETHOD
	 * )
	 * with parameters : 
	 * 	mbo is the workplan record (wplabor, wpmaterial, wptool, wpservice, workorder)
	 * 	job plan associated record (
	 * 		wplabor 	-> joblabor,
	 * 		wpmaterial 	-> jobmaterial,
	 * 		wptool 		-> jobtool,
	 * 		wpservice 	-> jobservice,
	 * 		workorder	-> jobplan
	 * 	)
	 * calculatedField : the names of the fields on the mbo and job plan associated records
	 * 					 that must be used to set and retrieve the values
	 * 					Example : if JOBRESOURCE.CALCAPPLYTO = HOURS, then we will return for WPLABOR the fields WPLABOR.LABORHRS and JOBLABOR.LABORHRS
	 * 					Example : if JOBRESOURCE.CALCAPPLYTO = HOURS, then we will return for WORKORDER the fields WORKORDER.ESTDUR and JOBPLAN.TASKDURATION
	 * 
	 * @throws MXApplicationException if the workorder is not WAPPR
	 */
	@Override
	public void calculateDynamicJobPlan() throws RemoteException, MXException {
		// Here we can calculate the dynamic job plan values only if the WO is wappr for approval
		MboSetRemote editSettings = getMboSet("$lookupSetttings", "WPEDITSETTING", "orgid = :orgid and status = '" + this.getInternalStatus() + "'") ;
		if ( editSettings.count() == 0 || !editSettings.getMbo(0).getBoolean("APPLYDJP")) {
			throw new MXApplicationException("dynamicjobplan","cannotCalculate");
		}
		
		calculateTotalWorkUnits();
		
		// Here, for each WPLABOR, WPMATERIAL, WPSERVICE and WPTOOL, we have to calculate the quantity or duration
		// according to the method
		MboSetRemote[] resources = {
				this.getMboSet(WOSetRemote.SHOWTASKS),
				this.getMboSet(WOSetRemote.SHOWPLANLABOR),
				this.getMboSet(WOSetRemote.SHOWPLANMATERIAL),
				this.getMboSet(WOSetRemote.SHOWPLANTOOL),
				this.getMboSet(WOSetRemote.SHOWPLANSERVICE)
				};
		
		
		try {
			calculateDynamicJobPlan(this, "WO");
			
			for(MboSetRemote mboSet : resources ) {
				//mboSet.reset();
				for(MboRemote mbo = mboSet.moveFirst(); mbo != null; mbo = mboSet.moveNext()) {
					calculateDynamicJobPlan(mbo, mbo.getThisMboSet().getMboSetInfo().getObjectName());
					
				}
			}
		
			setValue(DJPWO.DJPAPPLIED, "YES");
			setValueNull(DJPWO.DJPERROR);
			if ( getThisMboSet().hasWarnings()) {
				String msg = "";
				for(MXException e : getThisMboSet().getWarnings()) {
					if ( e.getErrorGroup().equals("dynamicjobplan")) {
						msg += e.getMessage(getThisMboSet()) + "\n";	
					}
				}
				setValue(DJPWO.DJPERROR, msg);
			}
//			boolean pmWorkOrderGen = MXServer.getBulletinBoard().isPosted(psdi.app.pm.PM.GENERATING_WORK);
//			if ( !pmWorkOrderGen ) {
//				getThisMboSet().addWarning(new MXApplicationException("dynamicjobplan","calculationSuccessfullyApplied"));
//			}
		} catch (MXException e) {
			setValue(DJPWO.DJPAPPLIED, "ERROR");
			setValue(DJPWO.DJPERROR, e.getMessage(getThisMboSet()));
		}
	}
	
	/**
	 * @see#calculateDynamicJobPlan()
	 * @param mbo the resource for which the calculation must be performed
	 * @param objectName the name of the object. Should be obtained by the getMboSetInfo().getObjectName() method, but could be overriden
	 * This name is used to get the fields in mboReferences, hoursFields and quantityFields maps.
	 * @throws RemoteException
	 * @throws MXException
	 */
	private void calculateDynamicJobPlan(MboRemote mbo, String objectName) throws RemoteException, MXException {
		MboRemote jobMboReference = getJobPlanMboReference(mbo, objectName);	// Could be JOBTASK, JOBLABOR, JOBITEM
		if ( jobMboReference == null ) return;
		
		String[] calculatedField = null;//this.getCalculatedFieldName(jobMboReference, objectName);
		boolean isHour = false;
		boolean isBoth = false;
		
		switch( mbo.getClass().equals(getClass()) ? CalcApplyTo.HOURS : DynamicJobPlanHelper.getApplyTo(jobMboReference) ) {
		case HOURS:
			calculatedField = hoursFields.get(objectName);
			isHour = true;
			break;
		case BOTH:
			isBoth = true;
		case QUANTITY:
			calculatedField = quantityFields.get(objectName);
			isHour = false;
			break;
		}
		
		if ( calculatedField == null ) return;	// Maybe we should raise an exception cause the data are not valid

		// This block is executed in priority with the QUANTITY parameter (if the jobplan must apply to BOTH quantity and hours)
		// Then, if must apply to BOTH the second block will be applied only with HOURS param
		String expression;
		if ( jobMboReference.getString(DJPConstants.CALCMETHOD).equals(DJPConstants.CALCMETHOD_PROPORTIONAL)) {
			calculateProportional(mbo, jobMboReference, calculatedField);
		} else if ( jobMboReference.getString(DJPConstants.CALCMETHOD).equals(DJPConstants.CALCMETHOD_LEVEL)) {
			calculateLevel(mbo, jobMboReference, calculatedField, isHour);
		} else if ( jobMboReference.getString(DJPConstants.CALCMETHOD).equals(DJPConstants.CALCMETHOD_PROPORTIONAL_ATTRIBUTE)) {
			expression = jobMboReference.getString(isHour && isBoth ? DJPConstants.CALCEXPRESSIONHOURS : DJPConstants.CALCEXPRESSION);
			
			calculateProportionalExpression(mbo, jobMboReference, calculatedField, expression);
		} else if ( jobMboReference.getString(DJPConstants.CALCMETHOD).equals(DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE)) {
			expression = jobMboReference.getString(isHour && isBoth ? DJPConstants.CALCEXPRESSIONHOURS : DJPConstants.CALCEXPRESSION) ;
			calculateLevelAttribute(mbo, jobMboReference, calculatedField, expression, isHour);
		}
		
		if (  isBoth && mbo.getClass().equals(getClass()) || DynamicJobPlanHelper.getApplyTo(jobMboReference) == CalcApplyTo.BOTH ) {
			calculatedField = hoursFields.get(objectName);
			if ( calculatedField == null ) return;	// Maybe we should raise an exception cause the data are not valid
			
			if ( jobMboReference.getString(DJPConstants.CALCMETHOD).equals(DJPConstants.CALCMETHOD_PROPORTIONAL)) {
				calculateProportional(mbo, jobMboReference, calculatedField);
			} else if ( jobMboReference.getString(DJPConstants.CALCMETHOD).equals(DJPConstants.CALCMETHOD_LEVEL)) {
				calculateLevel(mbo, jobMboReference, calculatedField, true);
			} else if ( jobMboReference.getString(DJPConstants.CALCMETHOD).equals(DJPConstants.CALCMETHOD_PROPORTIONAL_ATTRIBUTE)) {
				expression = jobMboReference.getString(DJPConstants.CALCEXPRESSIONHOURS);
				calculateProportionalExpression(mbo, jobMboReference, calculatedField, expression);
			} else if ( jobMboReference.getString(DJPConstants.CALCMETHOD).equals(DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE)) {
				expression = jobMboReference.getString(DJPConstants.CALCEXPRESSIONHOURS);
				calculateLevelAttribute(mbo, jobMboReference, calculatedField, expression, true);
			}
		}
	}
	
	
	private double calculateLevelAttributeForMultiAsset(MboRemote multiAsset, MboRemote plan, MboRemote jobres, String[] calculatedField, String expression, boolean isHour) throws RemoteException, MXException {
		if ( multiAsset.toBeDeleted() ) return 0;
		AssetRemote asset = (AssetRemote) multiAsset.getMboSet("ASSET").getMbo(0);
		if ( asset == null ) return 0;
		
		if ( asset.getBoolean("ISLINEAR") && mustApplyLinearCalculation()) {
			return calculateProportionalExpressionOnLinearAsset(multiAsset, asset, jobres, calculatedField[1], expression);
		} else {
			double totalWorkUnits = multiAsset.getDouble(TOTALWORKUNITS);
			return calculateProportionalExpressionOnNonLinearAsset(asset, jobres, calculatedField[1], expression, totalWorkUnits);
		}
	}
	
	/**
	 * 
	 * @param plan	
	 * 		the workplan record (could be WPLABOR, WPMATERIAL, WPTOOL, WPSERVICE, WORKORDER)
	 * @param 
	 * 		jobres the job plan associated records (could be JOBLABOR, JOBMATERIAL, JOBTOOL, JOBSERVICE, JOBPLAN)
	 * @param 
	 * 		calculatedField the names of the fields on the mbo and job plan associated records
	 * 
	 * @see DJPWO#calculateDynamicJobPlan()
	 * 
	 * @throws RemoteException
	 * @throws MXException
	 * 		for now it throw a NotYetImplemented Exception
	 */
	private void calculateLevelAttribute(MboRemote plan, MboRemote jobres, String[] calculatedField, String expression, boolean isHour) throws RemoteException, MXException {
		// for each linear asset
		MboSetRemote multiAssetLocCISet = getMboSet("MULTIASSETLOCCI");
//		multiAssetLocCISet.reset();
//		MboSetRemote multiAssetLocCISet = getAllMultiAssetLocCI();
		multiAssetLocCISet.setWhere("assetnum is not null");
		
		ResourceLevelsSetRemote resourceLevels = (ResourceLevelsSetRemote) jobres.getMboSet("RESLEVEL");
		resourceLevels.reset();
		ResourceLevelsRemote resourceLevel = (ResourceLevelsRemote) resourceLevels.getMbo(0);
		String unitOfResourceLevels = resourceLevel.getString(ResourceLevelsRemote.MEASUREUNITID);
		
		double totalResourceQty = 0;
		
		for(MboRemote multiAsset = multiAssetLocCISet.moveFirst(); multiAsset != null; multiAsset = multiAssetLocCISet.moveNext()) {
			if ( multiAsset.toBeDeleted() ) continue;
			double tmp = calculateLevelAttributeForMultiAsset(multiAsset, plan, jobres, calculatedField, expression, isHour);
			if ( !isHour ) {
				tmp = DynamicJobPlanHelper.convertValue(this, totalResourceQty, getString(DJPConstants.UNITSOFWORK), unitOfResourceLevels);
			}
			ResourceByLevelRemote resourceByLevel = resourceLevel.getLevelForValue(tmp);
			
			if (resourceByLevel == null) {
				throw new MXApplicationException("dynamicjobplan","invalidResourceLevel", new Object[]{totalResourceQty,resourceLevel.getResourceLevelsNum()});
			}
			totalResourceQty += isHour ? resourceByLevel.getHours() : resourceByLevel.getQuantity();
//			totalResourceQty += calculateLevelAttributeForMultiAsset(multiAsset, plan, jobres, calculatedField, expression, isHour);
		}
		multiAssetLocCISet.close();

		if ( getMboSet("LINEARRELATED").getMbo(0) != null ) {
			if ( !getMboSet("LINEARRELATED").getMbo(0).toBeDeleted() ) {
				double tmp = calculateLevelAttributeForMultiAsset(getMboSet("LINEARRELATED").getMbo(0), plan, jobres, calculatedField, expression, isHour);
				tmp = DynamicJobPlanHelper.convertValue(this, totalResourceQty, getString(DJPConstants.UNITSOFWORK), unitOfResourceLevels);
				ResourceByLevelRemote resourceByLevel = resourceLevel.getLevelForValue(tmp);
				
				if (resourceByLevel == null) {
					throw new MXApplicationException("dynamicjobplan","invalidResourceLevel", new Object[]{totalResourceQty,resourceLevel.getResourceLevelsNum()});
				}
				totalResourceQty += isHour ? resourceByLevel.getHours() : resourceByLevel.getQuantity();
			}
//			totalResourceQty += calculateLevelAttributeForMultiAsset(getMboSet("LINEARRELATED").getMbo(0), plan, jobres, calculatedField, expression, isHour);
		}
		
		
//		ResourceLevelsRemote resourceLevels = (ResourceLevelsRemote) jobres.getMboSet("RESLEVEL").getMbo(0);
//		String unitOfResourceLevels = resourceLevels.getString(ResourceLevelsRemote.MEASUREUNITID);
//		
//		if( myLogger.isInfoEnabled() ) {
//			myLogger.info("Calculate Level value for " + plan.getThisMboSet().getMboSetInfo().getObjectName() + " " + plan.getUniqueIDValue() + " with Level " + resourceLevels.getString(ResourceLevelsRemote.RESOURCELEVELSNUM));
//		}
		
//		if ( unitOfResourceLevels != null && !unitOfResourceLevels.isEmpty() ) {
//			totalResourceQty = DynamicJobPlanHelper.convertValue(this, totalResourceQty, getString(DJPConstants.UNITSOFWORK), unitOfResourceLevels);
//		}
//		
//		if( myLogger.isInfoEnabled() ) {
//			myLogger.info("\tvalue = " + totalResourceQty);
//		}
//		
//		ResourceByLevelRemote resourceByLevel = (resourceLevels).getLevelForValue(totalResourceQty);
//		
//		if (resourceByLevel == null) {
//			throw new MXApplicationException("dynamicjobplan","invalidResourceLevel", new Object[]{totalResourceQty,resourceLevels.getResourceLevelsNum()});
//		}
		
		plan.setValue(calculatedField[0], totalResourceQty);
//		if ( isHour ) {
//			plan.setValue(calculatedField[0], resourceByLevel.getHours());
//		} else {
//			plan.setValue(calculatedField[0], resourceByLevel.getQuantity());
//		}
	}

	
	/**
	 *  According to the MAXVAR "DYNAMIC_ERROR", which could be CONTINUE or STOP
	 * @return true is the maxvar value is STOP, false otherwise
	 * @throws RemoteException
	 * @throws MXException
	 */
	private boolean stopOnError() throws RemoteException, MXException {
		return "STOP".equalsIgnoreCase(getMboServer().getMaxVar().getString("DYNAMIC_ONERROR", getOrgSiteForMaxvar("DYNAMIC_ONERROR")));
	}
	
	private double calculateProportionalExpressionOnLinearAsset(MboRemote multiAsset, AssetRemote asset, MboRemote jobres, String calculatedField, String expression) throws RemoteException, MXException {
		String unitsOfWork = getString(DJPConstants.UNITSOFWORK);
		double jobplanUnitOfWork = DynamicJobPlanHelper.getJobPlan(jobres).getDouble(DJPConstants.UNITSOFWORKMULTIPLIER);
		jobplanUnitOfWork = DynamicJobPlanHelper.convertValue(this, jobplanUnitOfWork, DynamicJobPlanHelper.getJobPlan(jobres).getString(DJPConstants.UNITSOFWORK), unitsOfWork);
		
		// Get the workorder start measure and end measure
		double woStartMeasure = multiAsset.getDouble("STARTMEASURE");
		double woEndMeasure = multiAsset.getDouble("ENDMEASURE");
		String startMeasureUnitId = multiAsset.getString("STARTMEASUREUNITID");
		String endMeasureUnitId = multiAsset.getString("ENDMEASUREUNITID");
		
		woStartMeasure = DynamicJobPlanHelper.convertValue(this, woStartMeasure, startMeasureUnitId, unitsOfWork);
		woEndMeasure = DynamicJobPlanHelper.convertValue(this, woEndMeasure, endMeasureUnitId, unitsOfWork);
		
		float resourceQty = 0;
		Double res = null;
//		String expression = jobres.getString(DJPConstants.CALCEXPRESSION);
		PreparedScript<Double> script = new PreparedScript<Double>(expression);	
		List<LinearSegment> segments = PreparedScriptHelper.getSegments(script, asset, woStartMeasure, woEndMeasure, unitsOfWork);
		
		for(LinearSegment segment : segments) {
			if ( segment.getStart() >= woEndMeasure || segment.getEnd() <= woStartMeasure ) continue;
			script.clearParameterValues();
			
//			script.setParameterValue(DJPConstants.HOURS_TAG, jobres.getDouble(calculatedField));
//			script.setParameterValue(DJPConstants.QUANTITY_TAG, jobres.getDouble(calculatedField));

			for(String param : script.getParametersName()) {
//				if ( param.equals(DJPConstants.HOURS_TAG) || param.equals(DJPConstants.QUANTITY_TAG) ) continue;
				if ( segment.attributeNames().contains(param) ) {
					script.setParameterValue(param, segment.getAttribute(param));
				} else {
					MXException ex = new MXApplicationException("dynamicjobplan","attributeMissing",new Object[]{asset.getString("assetnum"), param, segment.getStart(), segment.getEnd()});
					if ( stopOnError() ) {
						throw ex; 
					} else {
						getThisMboSet().addWarning(ex);
					}
						
				}
			}
			
			try {
				double tmpStart = Math.max(segment.getStart(),woStartMeasure);
				double tmpEnd = Math.min(segment.getEnd(), woEndMeasure);
				double distance = Math.max(tmpStart, tmpEnd) - Math.min(tmpStart, tmpEnd);
				double jobplanQty = jobres.getDouble(calculatedField);
				res = jobplanQty * (Double)script.eval() * (distance) / jobplanUnitOfWork;
				resourceQty += res;
			} catch (ScriptException e) {
				if ( stopOnError() )
					throw new MXApplicationException("dynamicjobplan","invalidExpression",e);
			}
			
		}
		
		return resourceQty;
	}
	
	private double calculateProportionalExpressionOnNonLinearAsset(AssetRemote asset, MboRemote jobres, String calculatedField, String expression, double totalWorkUnits) throws RemoteException, MXException {
		String unitsOfWork = getString(DJPConstants.UNITSOFWORK);
//		double totalWorkUnits = getDouble(TOTALWORKUNITS);
		double jobplanUnitOfWork = DynamicJobPlanHelper.getJobPlan(jobres).getDouble(DJPConstants.UNITSOFWORKMULTIPLIER);
		jobplanUnitOfWork = DynamicJobPlanHelper.convertValue(this, jobplanUnitOfWork, DynamicJobPlanHelper.getJobPlan(jobres).getString(DJPConstants.UNITSOFWORK), unitsOfWork);
		
//		String expression = jobres.getString(DJPConstants.CALCEXPRESSION);
		Double res = null;
		
		// Here we should not set the value for both, but the Adapter on the JOBRESOURCE.CALCEXPRESSION should catch a wrong expression
		// i.e an expression that should not use one of HOURS or QUANTITY TAG
		PreparedScript<Double> script = new PreparedScript<Double>(expression);
		
		String isValid = PreparedScriptHelper.verifySegmentAttributesAreUnique(asset, script.getParametersName());
		if ( isValid != null ) {
			throw new MXApplicationException("dynamicjobplan","segmentHasValues", new Object[]{asset.getString("ASSETNUM"), asset.getString("SITEID"), isValid});
		}

		
		LinearSegment segment = PreparedScriptHelper.getSegment(script, asset);
		
//		script.setParameterValue(DJPConstants.HOURS_TAG, jobres.getDouble(calculatedField));
//		script.setParameterValue(DJPConstants.QUANTITY_TAG, jobres.getDouble(calculatedField));
		
		for(String param : script.getParametersName()) {
//			if ( param.equals(DJPConstants.HOURS_TAG) || param.equals(DJPConstants.QUANTITY_TAG) ) continue;
			if ( segment != null && segment.getAttribute(param) != null ) {
				script.setParameterValue(param, segment.getAttribute(param));
			} else {
				MXException ex = new MXApplicationException("dynamicjobplan","attributeMissing",new Object[]{asset.getString("assetnum"), param});
				if ( stopOnError() ) {
					throw ex; 
				} else {
					getThisMboSet().addWarning(ex);
				}
			}
		}
		
		try {
			res = (Double)script.eval();
			res = res * jobres.getDouble(calculatedField);
			res = res * totalWorkUnits / jobplanUnitOfWork;
		} catch (ScriptException e) {
			if ( stopOnError())
				throw new MXApplicationException("dynamicjobplan","invalidExpression",e);
		}
		
		return res;
	}
	
	
	/**
	 * Method use to detect if the job plan is set as dynamic and should be apply using the length of the 
	 * workorder (startmeasure and endmeasure) or not.
	 * If not, the TOTAL WORK UNITS of the jobresources will be use to calculate the results.
	 * @return true if the jobplan is dynamic and should apply the linear calculation. Otherwise return false
	 * @throws RemoteException
	 * @throws MXException
	 */
	public boolean mustApplyLinearCalculation() throws RemoteException, MXException {
		MboRemote jobplan = getJobPlanMboReference(null, "WO");
		return jobplan != null && !jobplan.getBoolean(DJPConstants.APPLYNONDYNAMICCALC);
	}
	
	
	private double calculateProportionalExpressionForMultiAsset(MboRemote multiAsset, MboRemote plan, MboRemote jobres, String[] calculatedField, String expression) throws RemoteException, MXException {
		if ( multiAsset.toBeDeleted() ) return 0;
		
		AssetRemote asset = (AssetRemote) multiAsset.getMboSet("ASSET").getMbo(0);
		if ( asset == null ) return 0;
		
		if ( asset.getBoolean("ISLINEAR") && mustApplyLinearCalculation() ) {
			return calculateProportionalExpressionOnLinearAsset(multiAsset, asset, jobres, calculatedField[1], expression);
		} else {
			double totalWorkUnits = multiAsset.getDouble(TOTALWORKUNITS);
			return calculateProportionalExpressionOnNonLinearAsset(asset, jobres, calculatedField[1], expression, totalWorkUnits);
		}
	}
	
	/**
	 * For each asset/location/CI apply the dynamic jobplan.
	 * @param plan
	 * @param jobres
	 * @param calculatedField
	 * @param expression
	 * @throws RemoteException
	 * @throws MXException
	 */
	private void calculateProportionalExpression(MboRemote plan, MboRemote jobres, String[] calculatedField, String expression) throws RemoteException, MXException {		
		// for each linear asset
//		MboSetRemote multiAssetLocCISet = getAllMultiAssetLocCI();
		MboSetRemote multiAssetLocCISet = getMboSet("MULTIASSETLOCCI");
//		multiAssetLocCISet.reset();
		multiAssetLocCISet.setWhere("assetnum is not null");
		
		double totalResourceQty = 0;
		
		for(MboRemote multiAsset = multiAssetLocCISet.moveFirst(); multiAsset != null; multiAsset = multiAssetLocCISet.moveNext()) {
			totalResourceQty += calculateProportionalExpressionForMultiAsset(multiAsset, plan, jobres, calculatedField, expression);
		}
		multiAssetLocCISet.close();
		if ( getMboSet("LINEARRELATED").getMbo(0) != null ) {
			totalResourceQty += calculateProportionalExpressionForMultiAsset(getMboSet("LINEARRELATED").getMbo(0), plan, jobres, calculatedField, expression);
		}
		
		plan.setValue(calculatedField[0], totalResourceQty);
	}

	private double calculateProportionalForMultiAsset(MboRemote multiAsset, MboRemote plan, MboRemote jobres, String[] calculatedField) throws RemoteException, MXException {
		if ( multiAsset.toBeDeleted() ) return 0;
		
		String unitsOfWork = getString(DJPConstants.UNITSOFWORK);
		double jobplanQty = jobres.getDouble(calculatedField[1]);
		double jobplanUnitOfWork = DynamicJobPlanHelper.getJobPlan(jobres).getDouble(DJPConstants.UNITSOFWORKMULTIPLIER);
		jobplanUnitOfWork = DynamicJobPlanHelper.convertValue(this, jobplanUnitOfWork, DynamicJobPlanHelper.getJobPlan(jobres).getString(DJPConstants.UNITSOFWORK), getString(DJPConstants.UNITSOFWORK));
		
		AssetRemote asset = (AssetRemote) multiAsset.getMboSet("ASSET").getMbo(0);

		if ( asset != null && asset.getBoolean("ISLINEAR") && mustApplyLinearCalculation() ) {
			double woStartMeasure = multiAsset.getDouble("STARTMEASURE");
			double woEndMeasure = multiAsset.getDouble("ENDMEASURE");
			String startMeasureUnitId = multiAsset.getString("STARTMEASUREUNITID");
			String endMeasureUnitId = multiAsset.getString("ENDMEASUREUNITID");
			
			woStartMeasure = DynamicJobPlanHelper.convertValue(this, woStartMeasure, startMeasureUnitId, unitsOfWork);
			woEndMeasure = DynamicJobPlanHelper.convertValue(this, woEndMeasure, endMeasureUnitId, unitsOfWork);
			
			return jobplanQty * Math.abs(woStartMeasure - woEndMeasure) / jobplanUnitOfWork;
		} else {
			
			
//			double totalWorkUnits = getDouble(TOTALWORKUNITS);
			double totalWorkUnits = multiAsset.getDouble(TOTALWORKUNITS);
			return jobplanQty * totalWorkUnits / jobplanUnitOfWork;
		}
	}
	
	private void calculateProportional(MboRemote plan, MboRemote jobres, String[] calculatedField) throws RemoteException, MXException {
		
		if( myLogger.isInfoEnabled() ) {
			myLogger.info("Calculate proportional value for " + plan.getThisMboSet().getMboSetInfo().getObjectName() + " " + plan.getUniqueIDValue() );
		}
		
		// for each linear asset
		MboSetRemote multiAssetLocCISet = getMboSet("MULTIASSETLOCCI");
		//multiAssetLocCISet.reset();
				
		double totalResourceQty = 0;
		boolean linearAsset = false;
		boolean nonLinearAsset = false;
		for(MboRemote multiAsset = multiAssetLocCISet.moveFirst(); multiAsset != null; multiAsset = multiAssetLocCISet.moveNext()) {
			AssetRemote asset = (AssetRemote) multiAsset.getMboSet("ASSET").getMbo(0);
			if ( asset != null && asset.getBoolean("ISLINEAR") && mustApplyLinearCalculation() ) {
				linearAsset = true;
				if ( nonLinearAsset ) {
					throw new MXApplicationException("dynamicjobplan","linearAndNonLinear");
				}
			} else {
				nonLinearAsset = true;
				if ( linearAsset ) {
					throw new MXApplicationException("dynamicjobplan","linearAndNonLinear");
				}
			}
			totalResourceQty += calculateProportionalForMultiAsset(multiAsset, plan, jobres, calculatedField);
		}
		multiAssetLocCISet.close();
		
		if ( getMboSet("LINEARRELATED").getMbo(0) != null ) {
			totalResourceQty += calculateProportionalForMultiAsset(getMboSet("LINEARRELATED").getMbo(0), plan, jobres, calculatedField);
		}
		
		plan.setValue(calculatedField[0], totalResourceQty);
		
		
	}
	
//	private double calculateLevelForMultiAsset(MboRemote multiAsset, MboRemote plan, MboRemote jobres, String[] calculatedField, String unitOfResourceLevels, boolean isHour) throws RemoteException, MXException {
//		if ( multiAsset.toBeDeleted() ) return 0;
//		
//		double totalWorkUnits = multiAsset.getDouble(TOTALWORKUNITS);
//		
//		
//		if( myLogger.isInfoEnabled() ) {
//			myLogger.info("Calculate Level value for " + plan.getThisMboSet().getMboSetInfo().getObjectName() + " " + plan.getUniqueIDValue());
//		}
//		
//		if ( unitOfResourceLevels != null && !unitOfResourceLevels.isEmpty() ) {
//			totalWorkUnits = DynamicJobPlanHelper.convertValue(this, totalWorkUnits, getString(DJPConstants.UNITSOFWORK), unitOfResourceLevels);
//		}
//		
//		if( myLogger.isInfoEnabled() ) {
//			myLogger.info("\tvalue = " + totalWorkUnits);
//		}
//		
//		return totalWorkUnits;
//	}
	
	private void calculateLevel(MboRemote plan, MboRemote jobres, String[] calculatedField, boolean isHour) throws RemoteException, MXException {
		
		ResourceLevelsSetRemote resourceLevels = (ResourceLevelsSetRemote) jobres.getMboSet("RESLEVEL");
		resourceLevels.reset();
		ResourceLevelsRemote resourceLevel = (ResourceLevelsRemote) resourceLevels.getMbo(0);
		String unitOfResourceLevels = resourceLevel.getString(ResourceLevelsRemote.MEASUREUNITID);
		
//		MboSetRemote multiAssetLocCISet = getAllMultiAssetLocCI();
		MboSetRemote multiAssetLocCISet = getMboSet("MULTIASSETLOCCI");
//		multiAssetLocCISet.reset();
		
		double calculatedResources = 0;
		for(MboRemote multiAsset = multiAssetLocCISet.moveFirst(); multiAsset != null; multiAsset = multiAssetLocCISet.moveNext()) {
			if ( multiAsset.toBeDeleted() ) continue;
//			double tmp = calculateLevelForMultiAsset(multiAsset, plan, jobres, calculatedField, unitOfResourceLevels, isHour);
			double tmp = DynamicJobPlanHelper.convertValue(this, multiAsset.getDouble(TOTALWORKUNITS), getString(DJPConstants.UNITSOFWORK), unitOfResourceLevels);
			ResourceByLevelRemote resourceByLevel = resourceLevel.getLevelForValue(tmp);
			if (resourceByLevel == null) {
				throw new MXApplicationException("dynamicjobplan","invalidResourceLevel");
			}
			calculatedResources += isHour ? resourceByLevel.getHours() : resourceByLevel.getQuantity();
			
//			calculatedResources += calculateLevelForMultiAsset(multiAsset, plan, jobres, calculatedField, unitOfResourceLevels, isHour);
		}
		
		if ( getMboSet("LINEARRELATED").getMbo(0) != null ) {
			if ( !getMboSet("LINEARRELATED").getMbo(0).toBeDeleted() ) {
//				double tmp = calculateLevelForMultiAsset(getMboSet("LINEARRELATED").getMbo(0), plan, jobres, calculatedField, unitOfResourceLevels, isHour);
				double tmp = DynamicJobPlanHelper.convertValue(this, getMboSet("LINEARRELATED").getMbo(0).getDouble(TOTALWORKUNITS), getString(DJPConstants.UNITSOFWORK), unitOfResourceLevels);
				ResourceByLevelRemote resourceByLevel = resourceLevel.getLevelForValue(tmp);
				if (resourceByLevel == null) {
					throw new MXApplicationException("dynamicjobplan","invalidResourceLevel");
				}
				calculatedResources += isHour ? resourceByLevel.getHours() : resourceByLevel.getQuantity();
//				calculatedResources += calculateLevelForMultiAsset(getMboSet("LINEARRELATED").getMbo(0), plan, jobres, calculatedField, unitOfResourceLevels, isHour);
			}
		}
		
		
//		ResourceByLevelRemote resourceByLevel = (resourceLevels).getLevelForValue(calculatedResources);		
		
		
		plan.setValue(calculatedField[0], calculatedResources);
//		if ( isHour ) {
//			plan.setValue(calculatedField[0], resourceByLevel.getHours());
//		} else {
//			plan.setValue(calculatedField[0], resourceByLevel.getQuantity());
//		}
		
		
		
	}

	/**
	 * According to the type of object (Aka OBJECTNAME) and the calculation method (Aka CALCMETHOD)
	 * The field that will be calculated will be a quantity field, or a duration field
	 * For Example : <br/>
	 * <ul>
	 * <li>for [JOBLABOR, QUANTITY method] -> Quantity</li>
	 * <li>for [JOBLABOR, HOURS method] -> LABORHRS</li>
	 * <li>for [JOBTOOL, QUANTITY method] -> ItemQty</li>
	 * <li>for [JOBTOOL, HOURS method] -> Hours</li>
	 * <li>...</li>
	 * </ul> 
	 * @param mbo the record that will be calculated. It could be a task/children (WOACTIVITY), a labor (WPLABOR), an item (WPMATERIAL,WPSERVICE, WPTOOL)
	 * @return a couple of value, representing the WORKPLAN destination field name, and JOBPLAN field name that is used as reference
	 * @throws RemoteException
	 * @throws MXException
	 */
//	private String[] getCalculatedFieldName(MboRemote mbo, String objectName) throws RemoteException, MXException {
//		// get the field name that will be calculated, according to the CALCAPPLYTO
//		String[] calculatedField = null;
//		switch(DynamicJobPlanHelper.getApplyTo(mbo)) {
//		case HOURS:
//			calculatedField = hoursFields.get(objectName);
//			break;
//		case QUANTITY:
//			calculatedField = quantityFields.get(objectName);
//			break;
//		default:
//			break;
//		
//		}
//		
//		return calculatedField;
//	}
	
	/**
	 * This method returns the corresponding jobplan record for resources/tasks in the workplan <br/>
	 * Ex : if mbo instanceof WPLABOR -> it returns the JOBLABOR record corresonding to mbo
	 * @param mbo a workplan record (WOACTIVITY/WORKORDER, WPLABOR, WPITEM, WPMATERIAL, WPSERVICE, WPTOOL)
	 * @return the corresponding record in the jobplan
	 * @throws RemoteException
	 * @throws MXException
	 */
	private MboRemote getJobPlanMboReference(MboRemote mbo, String objectName) throws RemoteException, MXException {
		JobPlanRemote jobPlan = (JobPlanRemote) getMboSet("$JOBPLAN","JOBPLAN","orgid = :orgid AND siteid = :siteid AND jpnum = :jpnum and :PLUSCJPREVNUM =  PLUSCREVNUM and status in (select value from synonymdomain where domainid = 'JOBPLANSTATUS' and maxvalue = 'ACTIVE')").getMbo(0);
		if ( jobPlan == null ) jobPlan = (JobPlanRemote) getMboSet("$JOBPLAN","JOBPLAN","orgid = :orgid AND jpnum = :jpnum and :PLUSCJPREVNUM =  PLUSCREVNUM and status in (select value from synonymdomain where domainid = 'JOBPLANSTATUS' and maxvalue = 'ACTIVE')").getMbo(0);
		if ( jobPlan == null ) jobPlan = (JobPlanRemote) getMboSet("$JOBPLAN","JOBPLAN","jpnum = :jpnum and :PLUSCJPREVNUM =  PLUSCREVNUM and status in (select value from synonymdomain where domainid = 'JOBPLANSTATUS' and maxvalue = 'ACTIVE')").getMbo(0);
		
		
		if ( jobPlan == null ) return null;
		if ( objectName.equalsIgnoreCase("WO") ) return jobPlan;
		
		MboSetRemote set = jobPlan.getMboSet(mboReferences.get(objectName));
		return set.getMboForUniqueId(mbo.getLong(set.getMboSetInfo().getUniqueIDName()));

	}
	
	@Override
	public void calculateTotalWorkUnitsFor(MboRemote multiAssetLocCI, String unitOfWork) throws RemoteException, MXException {
		Double endMeasure = DynamicJobPlanHelper.convertValue(this, multiAssetLocCI.getDouble("ENDMEASURE"),multiAssetLocCI.getString("ENDMEASUREUNITID"),unitOfWork);
		Double startMeasure = DynamicJobPlanHelper.convertValue(this, multiAssetLocCI.getDouble("STARTMEASURE"), multiAssetLocCI.getString("STARTMEASUREUNITID"),unitOfWork);
		multiAssetLocCI.setValue(DJPConstants.TOTALWORKUNITS, Math.abs( endMeasure - startMeasure ), MboConstants.NOACCESSCHECK);
	}
	
	/**
	 * Calculate the total length of attributes : for each linear asset in the multiassetlocci set,
	 * we add the absolute value for endmeasure - startmeasure (cause the endmeasure is not necessarely 
	 * lower than the startmeasure)<br/>
	 * Update the field TOTALWORKUNITS<br/>
	 * If some assets are not linear, they are skipped, and a warning is added to the MULTIASSETLOCCI set
	 */
	@Override
	public void calculateTotalWorkUnits() throws RemoteException, MXException {
	
		if ( myLogger.isInfoEnabled() ) {
			myLogger.info("Recalculating Total Work Units based on MULTIASSETLOCCI relationship on assets (assetnum not null)");
		}
		
		if ( !isNull("JPNUM") && isNull(DJPConstants.UNITSOFWORK) ) {
			setValue(DJPConstants.UNITSOFWORK, getString("JOBPLAN.UNITSOFWORK"), NOACCESSCHECK | NOACTION);
		}
		
		if ( !mustApplyLinearCalculation() ) {
			if ( myLogger.isInfoEnabled() ) {
				myLogger.info("Ignoring recalulation of total work units because the jobplan must not be applied as linear");
			}
			return;
		}
		
		MboSetRemote multiAssetLocCISet = getMboSet("MULTIASSETLOCCI");
//		multiAssetLocCISet.resetQbe();
		
		// First we take the linear segment details of the workorder
		MboRemote linear = getMboSet("LINEARRELATED").getMbo(0);
//		
//		
//		if ( linear != null && isNull(DJPConstants.UNITSOFWORK )) {
//			setValue(DJPConstants.UNITSOFWORK,linear.getString("STARTMEASUREUNITID"));
//		}
//		
//		if ( isNull(DJPConstants.UNITSOFWORK) && multiAssetLocCISet.count() >= 1) {
//			setValue(DJPConstants.UNITSOFWORK,multiAssetLocCISet.getMbo(0).getString("STARTMEASUREUNITID"));
//		}
		
		String unitOfWork = getString(DJPConstants.UNITSOFWORK);
		
		if ( linear != null && linear.getMboSet("ASSET").count() != 0 && linear.getBoolean("ASSET.ISLINEAR") ) {
			calculateTotalWorkUnitsFor(linear, unitOfWork);
		}

		 
		if ( multiAssetLocCISet.count() >= 1 ) {
			for(MboRemote multi = multiAssetLocCISet.moveFirst(); multi != null; multi = multiAssetLocCISet.moveNext()) {
				if ( multi.toBeDeleted()) continue;
				MboRemote asset = multi.getMboSet("ASSET").getMbo(0);
				
				if ( asset != null && asset.getBoolean("ISLINEAR")) {
					calculateTotalWorkUnitsFor(multi, unitOfWork);
				}
			}
		}
		
		setModified(true);
	}

	@Override
	public boolean isCalculationAppliedSuccessfully() throws RemoteException, MXException {
		return getString(DJPWO.DJPAPPLIED).equalsIgnoreCase("YES");
	}
	
	
	@Override
	public void copyRouteStopsToMultiAsset(MboSetRemote routestopSetRemote)
			throws MXException, RemoteException {
		super.copyRouteStopsToMultiAsset(routestopSetRemote);
		
		MboRemote jobplan = getJobPlanMboReference(this, "WO"); 
		if ( jobplan != null && !jobplan.getBoolean(DJPConstants.DYNAMIC) ) return;
		
		MboSetRemote multiAssets = getMboSet("MULTIASSETLOCCI");
		
		for(MboRemote mbo = multiAssets.moveFirst(); mbo != null; mbo = multiAssets.moveNext()) {
			MboRemote routeStop = mbo.getMboSet("$ROUTE_STOP","ROUTE_STOP","assetlocsiteid = :siteid and route = :route and routestopid = :routestop").getMbo(0);
			mbo.setValue(DJPConstants.TOTALWORKUNITS, routeStop.getDouble(DJPConstants.TOTALWORKUNITS), NOACCESSCHECK);
			mustApplyDJP = true;
		}
	}
	
	
	@Override
	protected void save() throws MXException, RemoteException {
		super.save();
		MboSetRemote editSettings = getMboSet("$lookupSetttings", "WPEDITSETTING", "orgid = :orgid and status = '" + this.getInternalStatus() + "'") ;
		if ( editSettings.count() == 0 || !editSettings.getMbo(0).getBoolean("APPLYDJP")) {
			return;
		}
		
		UserInfo ui = getUserInfo();
		
		MboRemote jobplan = getJobPlanMboReference(this, "WO");
		boolean shouldAsk = jobplan != null && jobplan.getBoolean(DJPConstants.DYNAMIC);
		boolean pmWorkOrderGen = MXServer.getBulletinBoard().isPosted(psdi.app.pm.PM.GENERATING_WORK);
		
		if ( pmWorkOrderGen ) {
			mustApplyDJP = true;
		} else if ( ui.isInteractive() && shouldAsk ) {
			int resp = MXApplicationYesNoCancelException.getUserInput("djp_wo_save", MXServer.getMXServer(), ui);
			if ( resp == MXApplicationYesNoCancelException.NULL ) {
				throw new MXApplicationYesNoCancelException("djp_wo_save", "dynamicjobplan","applyDJPAndSaveWO");
			} else {
				mustApplyDJP = resp == MXApplicationYesNoCancelException.YES;
			}
		}
		if ( mustApplyDJP ) {
			//calculateTotalWorkUnits();
			
			
			calculateDynamicJobPlan();
		}
	}
	
	private void updateTWU() throws RemoteException, MXException {
		MboRemote jobplan = getJobPlanMboReference(this, "WO"); 
		if ( jobplan != null && !jobplan.getBoolean(DJPConstants.DYNAMIC) ) return;
		
		MboSetRemote multiAssets = getMboSet("LINEARRELATED");
		MboRemote multiAssetRecord = multiAssets.getMbo(0);
	
		if ( multiAssetRecord != null && !isNull("PMNUM")) {
			MboSetRemote pmSequence = getMboSet("PM.PMSEQUENCE");
			pmSequence.setWhere("jpnum = '" + jobplan.getString("JPNUM") + "'");
			double val;
			val = pmSequence.getMbo(0) != null ? pmSequence.getMbo(0).getDouble("TOTALWORKUNITS") : Double.parseDouble(getMboValue(DJPConstants.TOTALWORKUNITS).getDefault());	
			
			multiAssetRecord.setValue(DJPConstants.TOTALWORKUNITS, val, NOACCESSCHECK);
			mustApplyDJP = true;
		}
	}
	
	@Override
	public void setmultiassetrecord() throws MXException, RemoteException {
		super.setmultiassetrecord();
		updateTWU();
		
	}
	
	public void copyLinearAssetFieldsFromRouteStop(MboRemote toMbo, MboRemote routestop) throws RemoteException, MXException {
		super.copyLinearAssetFieldsFromRouteStop(toMbo, routestop);
		toMbo.setValue(DJPConstants.TOTALWORKUNITS, routestop.getDouble(DJPConstants.TOTALWORKUNITS), MboConstants.NOACCESSCHECK);
			
	}
	
	@Override
	public void setAttrFromWoGen(MboRemote woGen) throws RemoteException,
			MXException {
		super.setAttrFromWoGen(woGen);
		updateTWU();
	}
}