package com.ibm.swg.util;

import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Map;

import psdi.app.inventory.ConversionSetRemote;
import psdi.mbo.Mbo;
import psdi.mbo.MboRemote;
import psdi.mbo.SqlFormat;
import psdi.mbo.Translate;
import psdi.server.MXServer;
import psdi.util.MXApplicationException;
import psdi.util.MXException;
import psdi.util.logging.MXLogger;
import psdi.util.logging.MXLoggerFactory;

import com.ibm.swg.util.DJPConstants.CalcApplyTo;

public class DynamicJobPlanHelper {

	
	private static MXLogger myLogger = MXLoggerFactory.getLogger("maximo.application.WOTRACK");
	
	/**
	 * Convert the given value from the measure unit fromUnitId to the measure unit tounitId
	 * The converion record must exists. If not, it throw an exception
	 *  
	 * @param value
	 * 			the double value to convert
	 * @param fromUnitId
	 * 			the measureunitid in which the value is currently specified
	 * @param toUnitId
	 * 			the measureunitid in which the value should be converted to
	 * @return the convert value according to the conversion factor. Cannot be null
	 * @throws RemoteException
	 * @throws MXException
	 */
	public static double convertValue(MboRemote thisMbo, double value, String fromUnitId, String toUnitId) throws RemoteException, MXException {
		if ( fromUnitId.equalsIgnoreCase(toUnitId) ) return value;
		ConversionSetRemote conversions = (ConversionSetRemote)MXServer.getMXServer().getMboSet("CONVERSION", thisMbo.getUserInfo());
		SqlFormat sqf = new SqlFormat(thisMbo,"(FROMMEASUREUNIT = :1  AND TOMEASUREUNIT = :2) OR (FROMMEASUREUNIT = :2  AND TOMEASUREUNIT = :1)");
		sqf.setObject(1, "CONVERSION", "FROMMEASUREUNIT", fromUnitId);
		sqf.setObject(2, "CONVERSION", "FROMMEASUREUNIT", toUnitId);
		conversions.setWhere(sqf.format());
		
//		conversions.setQbeExactMatch(true);
//		conversions.setQbe("FROMMEASUREUNIT", fromUnitId);
//		conversions.setQbe("TOMEASUREUNIT", toUnitId);
		if ( conversions.count() == 0 ) {
			throw new MXApplicationException("dynamicjobplan","noConversion",new String[]{fromUnitId, toUnitId});
		}
		
		Double conversionFactor = null;
		for(MboRemote mbo = conversions.moveFirst(); mbo != null && conversionFactor == null; mbo = conversions.moveNext()) {
			if ( mbo.getString("FROMMEASUREUNIT").equalsIgnoreCase(fromUnitId) ) conversionFactor = mbo.getDouble("CONVERSION");
		}
		
		// no direct conversion found. We have to reverse the conversion factor
		if ( conversionFactor == null ) conversionFactor = 1.0 / conversions.getMbo(0).getDouble("CONVERSION");
		conversions.close();
		if ( myLogger.isDebugEnabled() ) {
			myLogger.debug("Converting " + value + " from " + fromUnitId + " to " + toUnitId + " : " + (value * conversionFactor) + " " + toUnitId);
		}
		
		return value * conversionFactor;
	}
	
	/*
	 * For a given record which is a JOBPLAN, JOBTASK, JOBLABOR, JOBMATERIAL, JOBTOOL or JOBSERVICE
	 * It return the JOBPLAN record (if it is already a jobplan then it return the same record)
	 */
	public static MboRemote getJobPlan(MboRemote jobResource) throws RemoteException {
		if ( jobResource.isBasedOn("JOBPLAN") ) return jobResource;
		// else
		return jobResource.getOwner();
	}
	
	public static CalcApplyTo getApplyTo(MboRemote jobResource) throws RemoteException, MXException {
		Translate t = ((Mbo)jobResource).getTranslator();
		String calcApplyTo = t.toInternalString("CALCAPPLYTO", jobResource.getString(DJPConstants.CALCAPPLYTO));
		
		return CalcApplyTo.fromStringTo(calcApplyTo);
	}
	
	
	/**
	 * this part is used to enable / disable the ability for the user to put a dummy value
	 * in the field hours or quantity on a job resource (joblabor, jobmaterial, etc)
	 * when he selected the BY LEVEL or BY LEVEL BY ATTRIBUTE calculation method
	 * .
	 */
	private static final Map<String,String> quantityFields = new HashMap<String,String>();
	private static final Map<String,String> hoursFields = new HashMap<String,String>();
	
	static {
		quantityFields.put("JOBLABOR", "QUANTITY");
		quantityFields.put("JOBTOOL", "ITEMQTY");
		quantityFields.put("JOBSERVICE", "ITEMQTY");
		quantityFields.put("JOBMATERIAL", "ITEMQTY");
		// quantityFields.put("JOBPLAN", new String[]{"",""});		// NO SENSE ! 
		
		hoursFields.put("JOBLABOR", "LABORHRS");
		hoursFields.put("JOBTOOL", "HOURS");
		hoursFields.put("JOBTASK", "TASKDURATION");
		// hoursFields.put("WPSERVICE", new String[]{"",""});			// NO SENSE !
		// hoursFields.put("WPMATERIAL", new String[]{"",""});			// NO SENSE !
		hoursFields.put("JOBPLAN", "JPDURATION");
	}
	
	/**
	 * this part is used to enable / disable the ability for the user to put a dummy value
	 * in the field hours or quantity on a job resource (joblabor, jobmaterial, etc)
	 * when he selected the BY LEVEL or BY LEVEL BY ATTRIBUTE calculation method.
	 * If the method is one of these two, then :
	 * - if calculating hours, the JOBRESOURCE.HOURS is not modifiable by the user (and set by default to 1, to have an accurate expression evaluation) 
	 * - if calculating quantity, the JOBRESOURCE.QUANTITY is not modifiable by the user (and set by default to 1, to have an accurate expression evaluation)
	 */
//	public static void onJobResourceMethodchanged(MboRemote jobResource) throws RemoteException, MXException {
//		String objectName = jobResource.getThisMboSet().getMboSetInfo().getObjectName();
//		String fieldName;	// field to set readonly
//		String fieldName2; // field to unset readonly;
//		Translate t = ((Mbo)jobResource).getTranslator();
//		String calcMethod = t.toInternalString("CALCMETHOD", jobResource.getString(DJPConstants.CALCMETHOD));
//		
//		if ( DJPConstants.CALCMETHOD_LEVEL.equalsIgnoreCase(calcMethod) || DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE.equalsIgnoreCase(calcMethod)) {
//			String calcApplyTo = t.toInternalString("CALCAPPLYTO", jobResource.getString(DJPConstants.CALCAPPLYTO));
//			if ( !jobResource.isFlagSet(MboConstants.READONLY) ) {
//				
//				switch(CalcApplyTo.fromStringTo(calcApplyTo)) {
//					case HOURS:
//						if ( hoursFields.get(objectName) != null && jobResource.getThisMboSet().getMboSetInfo().getAttribute(hoursFields.get(objectName)) != null ) {
//							jobResource.setValue(hoursFields.get(objectName), 1.0, MboConstants.NOACCESSCHECK);
//							jobResource.setFieldFlag(hoursFields.get(objectName), MboConstants.READONLY, true);
//						}
//						if ( quantityFields.get(objectName) != null && jobResource.getThisMboSet().getMboSetInfo().getAttribute(quantityFields.get(objectName)) != null ) {
//							jobResource.setFieldFlag(quantityFields.get(objectName), MboConstants.READONLY, false);
//						}
//						break;
//					case QUANTITY:
//						if ( quantityFields.get(objectName) != null && jobResource.getThisMboSet().getMboSetInfo().getAttribute(quantityFields.get(objectName)) != null ) {
//							jobResource.setValue(quantityFields.get(objectName), 1.0, MboConstants.NOACCESSCHECK);
//							jobResource.setFieldFlag(quantityFields.get(objectName), MboConstants.READONLY, true);
//						}
//						if ( hoursFields.get(objectName) != null && jobResource.getThisMboSet().getMboSetInfo().getAttribute(hoursFields.get(objectName)) != null ) {
//							jobResource.setFieldFlag(hoursFields.get(objectName), MboConstants.READONLY, false);		
//						}
//						
//						
//						break;
//					case BOTH:
//						jobResource.setValue(hoursFields.get(objectName), 1.0, MboConstants.NOACCESSCHECK);
//						jobResource.setFieldFlag(hoursFields.get(objectName), MboConstants.READONLY, true);
//						jobResource.setValue(quantityFields.get(objectName), 1.0, MboConstants.NOACCESSCHECK);
//						jobResource.setFieldFlag(quantityFields.get(objectName), MboConstants.READONLY, true);
//				}
//			}
//		} else {
//			// if both QUANTITY and HOURS fields are readonly, the jobplan is ACTIVE, or at least not modifiable
//			fieldName = quantityFields.get(objectName);
//			fieldName2 = hoursFields.get(objectName);
//			
//			if ( jobResource.isBasedOn("JOBLABOR") && !jobResource.isNull("LABORCODE")) {
//				fieldName = null;
//			}
//			if ( !jobResource.isFlagSet(MboConstants.READONLY) ) {
//				if ( fieldName != null ) {
//					jobResource.setFieldFlag(fieldName, MboConstants.READONLY, false);	
//				}
//				if ( fieldName2 != null ) {
//					jobResource.setFieldFlag(fieldName2, MboConstants.READONLY, false);
//				}
//			}
//			
//			
//			boolean flag = jobResource.getThisMboSet().getMboSetInfo().getAttribute(DJPConstants.CALCEXPRESSIONHOURS) != null && (jobResource.getString(DJPConstants.CALCMETHOD).equals(DJPConstants.CALCMETHOD_PROPORTIONAL_ATTRIBUTE) || jobResource.getString(DJPConstants.CALCMETHOD).equals(DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE)) && CalcApplyTo.fromStringTo(jobResource.getString(DJPConstants.CALCAPPLYTO)) == CalcApplyTo.BOTH;
//			jobResource.setFieldFlag(DJPConstants.CALCEXPRESSIONHOURS, MboConstants.REQUIRED, flag);
//		}
//	}
}
