package com.ibm.swg.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

public class LinearSegment /*implements Comparable<LinearSegment>*/ {

	
	private double start;
	
	private double end;
	
	private Map<String, Double> attributes;
	
	private LinearSegment(double start, double end) {
		if ( start > end ) {
			this.start = end;
			this.end = start;
		} else {
			this.start = start;
			this.end = end;
		}
		this.attributes = new HashMap<String, Double>();
	}
	
	public LinearSegment(double start, double end, String attributeName,
			double value) {
		this(start, end);
		this.attributes.put(attributeName, value);
	}

	public double getStart() {
		return start;
	}

	public double getEnd() {
		return end;
	}
	
	public Double getAttribute(String name) {
		return attributes.get(name);
	}
	
	public Set<String> attributeNames() {
		return attributes.keySet();
	}
	
	public boolean startBefore(LinearSegment linear) {
		return start < linear.start;
	}
	
	public boolean endBefore(LinearSegment linear) {
		return end < linear.end;
	}
	
	public boolean intersect(LinearSegment linear) {
		return  ((start > linear.start && start < linear.end) || (end > linear.start && end < linear.end)) ||
				((linear.start > start && linear.start < end) || (linear.end > start && linear.end < end)) ||
				( start == linear.start && end == linear.end );
	}
	
	public static LinearSegment intersection(LinearSegment l1, LinearSegment l2) {
		if ( !l1.intersect(l2)) return null;
		
		double s = Math.max(l1.start, l2.start);
		double e = Math.min(l1.end, l2.end);
		
		LinearSegment segment = new LinearSegment(s,e);
		segment.attributes.putAll(l1.attributes);
		segment.attributes.putAll(l2.attributes);
		return segment;
	}
		
	public static List<LinearSegment> substraction(LinearSegment l1, LinearSegment l2) {
		if ( !l1.intersect(l2)) return null;	// cas 1 et 2
		if ( l1.start == l2.start && l1.end == l2.end) return null;
		
		if ( l1.startBefore(l2) ) {
			if ( l1.endBefore(l2) ) {			// cas 3
				LinearSegment res = new LinearSegment(l1.start, l2.start);
				res.attributes.putAll(l1.attributes);
				return Arrays.asList(res);
			} else {							// cas 6
				LinearSegment res1 = new LinearSegment(l1.start, l2.start);
				LinearSegment res2 = new LinearSegment(l2.end, l1.end);
				res1.attributes.putAll(l1.attributes);
				res2.attributes.putAll(l1.attributes);
				return Arrays.asList(res1,res2);
			}
		} else {
			if ( l1.endBefore(l2)) {			// cas 5 
				return null;
			} else {							// cas 4
				LinearSegment res = new LinearSegment(l2.end, l1.end);
				res.attributes.putAll(l1.attributes);
				return Arrays.asList(res);
			}
		}
	}

	public static List<LinearSegment> aggregate(LinearSegment...segments) {
		return aggregate(Arrays.asList(segments));
	}
	
	public static List<LinearSegment> aggregate(List<LinearSegment> segments) {
		Set<Double> thresholds = new TreeSet<Double>();
		List<LinearSegment> res = new ArrayList<LinearSegment>();
		
		if ( segments.isEmpty() ) return res;
		
		for(LinearSegment segment : segments) {
			thresholds.add(segment.getStart());
			thresholds.add(segment.getEnd());
		}
		
		Iterator<Double> it = thresholds.iterator();
		Double start = it.next();
		while(it.hasNext()) {
			Double end = it.next();
			res.add(new LinearSegment(start, end));
			start = end;
		}
		
		for(LinearSegment segment : res) {
			for(LinearSegment tmp : segments) {
				if ( segment.intersect(tmp)) {
					segment.attributes.putAll(tmp.attributes);
				}
			}
		}
		return res;
	}
	
//	public static List<LinearSegment> aggregate(List<LinearSegment> segments) {
//		List<LinearSegment> orderedSegments = new ArrayList<LinearSegment>(segments);
//		List<LinearSegment> result = new ArrayList<LinearSegment>();
//		List<LinearSegment> tmp;
//		Collections.sort(orderedSegments);
//		
//		if ( orderedSegments.size() == 0 ) return result;
//		result.add(orderedSegments.get(0));
//		
//		for(LinearSegment current : orderedSegments) {
//			if ( current == orderedSegments.get(0)) continue;
//			List<LinearSegment> toAdd = new ArrayList<LinearSegment>();
//			LinearSegment toRemove = null;
//			for(LinearSegment ls : result) {
//				if ( ls.intersect(current)) {
//					toRemove = ls;
//					tmp = LinearSegment.substraction(ls, current);
//					if ( tmp != null ) toAdd.addAll(tmp);
//					toAdd.add(LinearSegment.intersection(current, ls));
//					 tmp = LinearSegment.substraction(current, ls);
//					if ( tmp != null ) toAdd.addAll(tmp);
//				}
//			}
//			
//			if ( toRemove != null ) {
//				result.remove(toRemove);
//			} else {
//				toAdd.add(current);
//			}
//			result.addAll(toAdd);
//		}
//		return result;
//		
//	}
	
//	@Override
//	public boolean equals(Object obj) {
//		return obj != null && obj instanceof LinearSegment && ((LinearSegment)obj).start == start && ((LinearSegment)obj).end == end ;
//	}
	
	@Override
	public String toString() {
		return String.format("%f - %f : %s", start, end, attributes);
	}

//	@Override
//	public int compareTo(LinearSegment o) {
//		return ((Double)start).compareTo((Double)o.start);
//	}
	
	
	public static void main(String[] args) {
		List<LinearSegment> segments = new ArrayList<LinearSegment>();
		
		segments.add(new LinearSegment(0.0, 54.21, "TRAVERSES:V_NBR_TR", 1666));
		segments.add(new LinearSegment(0.0, 54.21, "ATTACHES:V_NBRE", 2));
		segments.add(new LinearSegment(54.210000,54.320000, "ATTACHES:V_NBRE",3.0)); 
		segments.add(new LinearSegment(54.210000,134.600000, "TRAVERSES:V_NBR_TR",1722.0)); 
		segments.add(new LinearSegment(54.320000, 54.630000, "ATTACHES:V_NBRE", 2.0));
		segments.add(new LinearSegment(54.630000, 57.010000, "ATTACHES:V_NBRE", 2.0));
		segments.add(new LinearSegment(57.010000, 134.600000, "ATTACHES:V_NBRE", 3.0)); 
		segments.add(new LinearSegment(134.600000, 964.640000, "TRAVERSES:V_NBR_TR", 1666.0)); 
		segments.add(new LinearSegment(134.600000, 964.640000, "ATTACHES:V_NBRE", 2.0));

		for(LinearSegment seg : LinearSegment.aggregate(segments)) {
			System.out.println(seg);
		}
	}
}
