package com.ibm.swg.util;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;
import javax.script.ScriptException;

public class PreparedScript<T> {
	
	private String script;
	
	private Map<String, T> context;
	private Map<String, String> remapContext;
	
	public PreparedScript(String script) {
		context = new HashMap<String, T>();
		remapContext = new HashMap<String, String>();
		
		//context.put("$HOURS", null);
		//context.put("$QUANTITY", null);
		//remapContext.put("$QUANTITY", "$QUANTITY");
		//remapContext.put("$HOURS", "$HOURS");
		
		this.script = script;
		
		prepareScript();
	}
	
	private void prepareScript() {
		int i = 1;
		while ( script.indexOf(getStartDelimiter()) != -1 ) {
			int startIndex = script.indexOf(getStartDelimiter()) + 1;
			int endIndex = script.indexOf(getEndDelimiter());
			if ( endIndex < startIndex ) {	// no end delimiter => malformed script
				throw new IllegalArgumentException("No end delimiter after column " + (startIndex - 1));
			}
			String attrName = script.substring(startIndex, endIndex);
			
			remapContext.put(attrName.toUpperCase(), "_attr" + i);
			context.put(attrName.toUpperCase(), null);
			
			script = script.replace(getStartDelimiter() + attrName + getEndDelimiter(), "_attr" + i);
			i++;
		}		
	}
	
	public T eval() throws ScriptException {
		ScriptEngineManager mgr = new ScriptEngineManager();
	    ScriptEngine engine =  mgr.getEngineByName("JavaScript");
	    
		for(String s : remapContext.keySet()) {
			engine.put(remapContext.get(s), context.get(s));
		}
		
		@SuppressWarnings("unchecked")
		T res = (T) engine.eval(script);
		return res;
		
	}
	
	public void setParameterValue(String parameter, T value) {
		if ( !context.keySet().contains(parameter) ) {
			throw new IllegalArgumentException("This parameter is not used in the script");
		}
		context.put(parameter, value);
	}
	
	public Set<String> getParametersName() {
		return context.keySet();
	}
	
	public T getParameter(String key) {
		if ( !context.keySet().contains(key) ) {
			throw new IllegalArgumentException("This parameter is not used in the script");
		}
		return context.get(key);
	}
	
	public static char getStartDelimiter() {
		return '[';
	}

	public static char getEndDelimiter() {
		return ']';
	}

	public void clearParameterValues() {
		for(String s : getParametersName()) {
			setParameterValue(s, null);
		}
		
	}
	
	public boolean paramIsFeature(String paramName) {
		return context.keySet().contains(paramName) && paramName.contains(":");
	}
	
	public Set<String> getFeaturesNames() {
		Set<String> res = new HashSet<String>();
		
		for(String paramName : context.keySet()) {
			if ( paramName.contains(":")) {
				res.add(paramName.split(":")[0]);
			}
		}
		
		return res;
	}
	
	
	public Set<String> getFeatureParamNames(String featureName) {
		Set<String> res = new HashSet<String>();
		
		for(String paramName : context.keySet()) {
			if ( paramName.startsWith(featureName)) {
				res.add(paramName.split(":")[1]);
			}
		}
		
		return res;
	}

}

