package com.ibm.swg.util;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import psdi.mbo.MboRemote;
import psdi.mbo.MboSetRemote;
import psdi.mbo.SqlFormat;
import psdi.util.MXException;

public class PreparedScriptHelper {

	/**
	 * Check that the attributes have only 1 value for each segment.
	 * A case can come when we use linear asset with a dynamic jobplan and we do not want to
	 * use the linear referencement method.
	 * So for the calculation of the with attribute expressions, we must have only 1 value by attribute
	 * @param asset
	 * @param attributes
	 * @return
	 * @throws MXException 
	 * @throws RemoteException 
	 */
	public static String verifySegmentAttributesAreUnique(MboRemote asset, Set<String> attributes) throws RemoteException, MXException {
		Set<String> attributesCP = new HashSet<String>();
		Map<String, List<String>> attributesByFunction = new HashMap<String, List<String>>();
		StringBuilder results = new StringBuilder();
		
		for(String s : attributes ) {
			if ( !s.contains(":") ) {
				attributesCP.add(s);
			} else {
				String[] vals = s.split(":"); 
				if ( !attributesByFunction.containsKey(vals[0])) {
					attributesByFunction.put(vals[0].toUpperCase(), new ArrayList<String>());
				}
				attributesByFunction.get(vals[0].toUpperCase()).add(vals[1]);
			}
		}
		
		boolean verif = true;
		if ( !attributesCP.isEmpty() ) {
			MboSetRemote assetSpecs = asset.getMboSet("ASSETSPEC");
			StringBuilder sb = new StringBuilder();
			for(String attr : attributesCP) {
				sb.append("=").append(attr).append(",");
			}
			sb.deleteCharAt(sb.length() - 1);
			assetSpecs.setQbe("ASSETATTRID", sb.toString());
			verif = assetSpecs.count() == attributesCP.size();
			if ( !verif ) {
				for(MboRemote attr = assetSpecs.moveFirst(); attr != null; attr = assetSpecs.moveNext()) {
					results.append(attr.getString("ASSETATTRID")).append(", ");
				}
			}
			
			assetSpecs.close();
		}
		if ( verif && !attributesByFunction.isEmpty()) {
			MboSetRemote features = asset.getMboSet("ASSETFEATURES");
			StringBuilder sb = new StringBuilder();
			for(String feat : attributesByFunction.keySet()) {
				sb.append("=").append(feat).append(",");
			}
			sb.deleteCharAt(sb.length() - 1);
			features.setQbe("FEATURE", sb.toString());
			for(MboRemote feature = features.moveFirst(); verif && feature != null; feature = features.moveNext()) {
				List<String> assetAttributes = attributesByFunction.get(feature.getString("FEATURE"));
				sb = new StringBuilder();
				for(String attr : assetAttributes) {
					sb.append("=").append(attr).append(",");
				}
				MboSetRemote assetSpecs = feature.getMboSet("ASSETFEATURESPECCLASS");
				assetSpecs.setQbe("ASSETATTRID", sb.toString());
				verif = verif && assetSpecs.count() == assetAttributes.size();
				if ( !verif ) {
					for(MboRemote attr = assetSpecs.moveFirst(); attr != null; attr = assetSpecs.moveNext() ) {
						results.append(feature.getString("FEATURE")).append(":").append(attr.getString("ASSETATTRID")).append(", ");
					}
				}
				assetSpecs.close();
			}
			features.close();
		}
		
		if ( results.length() >= 2 ) {
			results.delete(results.length() - 2, results.length() - 1);
		}
		return verif ? null : results.toString();
	}
	
	/**
	 * @see PreparedScriptHelper#getSegments(PreparedScript, MboRemote, Double, Double)
	 * @param script
	 * @param asset
	 * @return only one 
	 * @throws RemoteException
	 * @throws MXException
	 */
	public static <T> LinearSegment getSegment(PreparedScript<T> script, MboRemote asset) throws RemoteException, MXException {
		List<LinearSegment> res = getSegments(script, asset, null, null, null);
		return res == null || res.isEmpty() ? null : res.get(0);
	}
	
	/**
	 * Get LinearSegments for a linear asset
	 * @param script The script that list the attributes / functions used by the expression
	 * @param asset The linear asset which hold the linear attributes
	 * @param woStartMeasure the start measure of the WO (multiassetlocci start or simulation start)  in which the calculation must be performed. if null the linear segments will be return as a segment (0-1) 
	 * @param woEndMeasure the end measure of the WO (multiassetlocci end or simulation end) in which the calculation must be performed. if null the linear segments will be return as a segment (0-1)
	 * @return an ORDERED collection representing the attribute values by segments for the script
	 */
	public static <T> List<LinearSegment> getSegments(PreparedScript<T> script, MboRemote asset, Double woStartMeasure, Double woEndMeasure, String unitMeasure) throws RemoteException, MXException {
		boolean mustUseMeasure = woStartMeasure != null && woEndMeasure != null;
		
//		SqlFormat sqf;
		StringBuilder sqfWhere;
		List<LinearSegment> segments = new ArrayList<LinearSegment>();
		
		// Loading asset attributes parameter values
		MboSetRemote assetSpecs = asset.getMboSet("ASSETSPEC");
		
		sqfWhere = new StringBuilder("");
		//if ( mustUseMeasure ) {
		//	sqfWhere.append("endmeasure >= :1 and startmeasure <= :2 and ");
		//}
		sqfWhere.append("assetattrid in ( ");
		
		boolean assetAttributeFound = false;
		for(String p : script.getParametersName()) {
//			if ( !script.paramIsFeature(p) && !p.equals(DJPConstants.HOURS_TAG) && !p.equals(DJPConstants.QUANTITY_TAG)) {
			if ( !script.paramIsFeature(p) ) {
				assetAttributeFound = true;
				sqfWhere.append("'").append(SqlFormat.getSQLString(p)).append("',");
			}
				
		}
		
		if ( assetAttributeFound ) {
			sqfWhere.deleteCharAt(sqfWhere.length()-1);
			sqfWhere.append(')');

			assetSpecs.setWhere(sqfWhere.toString());
			assetSpecs.setOrderBy("STARTMEASURE, ENDMEASURE asc");
			
			for(MboRemote attr = assetSpecs.moveFirst(); attr != null; attr = assetSpecs.moveNext()) {
				double start = 0;
				double end = 1;
				if ( mustUseMeasure && !attr.isNull("STARTMEASURE") && !attr.isNull("ENDMEASURE") ) {
					start = attr.getDouble("STARTMEASURE");
					end = attr.getDouble("ENDMEASURE");
					String startMeasureUnitId = attr.getString("STARTMEASUREUNITID");
					String endMeasureUnitId = attr.getString("ENDMEASUREUNITID");
					end = DynamicJobPlanHelper.convertValue(attr, end, endMeasureUnitId, unitMeasure);
					start = DynamicJobPlanHelper.convertValue(attr, start, startMeasureUnitId, unitMeasure);
					
					if ( end < woStartMeasure || start > woEndMeasure ) {
						continue;
					}
				}
				segments.add(new LinearSegment( start, end, attr.getString("ASSETATTRID"), attr.getDouble("NUMVALUE")));
			}
			assetSpecs.close();
		}
		// Loading features parameter values
		if ( !script.getFeaturesNames().isEmpty() ) {
			sqfWhere = new StringBuilder("(refobjectid, classstructureid) IN (select assetfeatureid,classstructureid from assetfeature where assetnum = :assetnum and siteid = :siteid ) and refobjectname = 'ASSETFEATURE' ");
//			if ( mustUseMeasure ) {
//				sqfWhere.append(" and endmeasure >= :1 and startmeasure <= :2");
//			}
			sqfWhere.append(" and (feature,assetattrid) in ( ");
			for(String featureName : script.getFeaturesNames()) {
				for(String featureParamName : script.getFeatureParamNames(featureName)) {
					sqfWhere.append("('").append(featureName.toUpperCase()).append("','").append(featureParamName).append("'),");
					
				}
			}
			sqfWhere.deleteCharAt(sqfWhere.length()-1);
			sqfWhere.append(')');
			
//			sqf = new SqlFormat(asset, sqfWhere.toString());
//			if ( mustUseMeasure ) {
//				sqf.setDouble(1, woStartMeasure);
//				sqf.setDouble(2, woEndMeasure);
//			}
//			
//			MboSetRemote featuresAttrSet = asset.getMboSet("$FEATUREATTR","ASSETFEATURESPEC",sqf.format());
			MboSetRemote featuresAttrSet = asset.getMboSet("$FEATUREATTR","ASSETFEATURESPEC", sqfWhere.toString());
			
			featuresAttrSet.setOrderBy("STARTMEASURE, ENDMEASURE asc");
			for(MboRemote featureAttr = featuresAttrSet.moveFirst(); featureAttr != null; featureAttr = featuresAttrSet.moveNext()) {
				double start = 0;
				double end = 1;
				if ( mustUseMeasure && !featureAttr.isNull("STARTMEASURE") && !featureAttr.isNull("ENDMEASURE") ) {
					start = featureAttr.getDouble("STARTMEASURE");
					end = featureAttr.getDouble("ENDMEASURE");
					String startMeasureUnitId = featureAttr.getString("STARTMEASUREUNITID");
					String endMeasureUnitId = featureAttr.getString("ENDMEASUREUNITID");
					end = DynamicJobPlanHelper.convertValue(featureAttr, end, endMeasureUnitId, unitMeasure);
					start = DynamicJobPlanHelper.convertValue(featureAttr, start, startMeasureUnitId, unitMeasure);
					
					if ( end < woStartMeasure || start > woEndMeasure ) {
						continue;
					}
				}
				segments.add(new LinearSegment(start, end, featureAttr.getString("FEATURE") + ":" + featureAttr.getString("ASSETATTRID"), featureAttr.getDouble("NUMVALUE")));
			}
			featuresAttrSet.close();
		}
		segments = LinearSegment.aggregate(segments);
//		Collections.sort(segments);
		
		return segments;
	}
	
	/*
	 * Get LinearSegments for a linear asset
	 * @param script The script that list the attributes / functions used by the expression
	 * @param asset The linear asset which hold the linear attributes
	 * @param woStartMeasure the start measure of the WO (multiassetlocci start or simulation start)  in which the calculation must be performed. if null the linear segments will be return as a segment (0-1) 
	 * @param woEndMeasure the end measure of the WO (multiassetlocci end or simulation end) in which the calculation must be performed. if null the linear segments will be return as a segment (0-1)
	 * @return an ORDERED collection representing the attribute values by segments for the script
	public static <T> List<LinearSegment> getSegments(PreparedScript<T> script, MboRemote asset, Double woStartMeasure, Double woEndMeasure, String unitMeasure) throws RemoteException, MXException {
		boolean mustUseMeasure = woStartMeasure != null && woEndMeasure != null;
		
		SqlFormat sqf;
		StringBuilder sqfWhere;
		List<LinearSegment> segments = new ArrayList<LinearSegment>();
		
		// Loading asset attributes parameter values
		MboSetRemote assetSpecs = asset.getMboSet("ASSETSPEC");
		
		sqfWhere = new StringBuilder("");
		if ( mustUseMeasure ) {
			sqfWhere.append("endmeasure >= :1 and startmeasure <= :2 and ");
		}
		sqfWhere.append("assetattrid in ( ");
		
		boolean assetAttributeFound = false;
		for(String p : script.getParametersName()) {
//			if ( !script.paramIsFeature(p) && !p.equals(DJPConstants.HOURS_TAG) && !p.equals(DJPConstants.QUANTITY_TAG)) {
			if ( !script.paramIsFeature(p) ) {
				assetAttributeFound = true;
				sqfWhere.append('\'').append(SqlFormat.getSQLString(p)).append("',");
			}
				
		}
		sqfWhere.deleteCharAt(sqfWhere.length()-1);
		sqfWhere.append(')');
		
		if ( assetAttributeFound ) {
			sqf = new SqlFormat(asset.getUserInfo(), sqfWhere.toString());
			
			if ( mustUseMeasure ) {
				sqf.setDouble(1, woStartMeasure);
				sqf.setDouble(2, woEndMeasure);
			}
	
			assetSpecs.setWhere(sqf.format());
			assetSpecs.setOrderBy("STARTMEASURE, ENDMEASURE asc");
			
			for(MboRemote attr = assetSpecs.moveFirst(); attr != null; attr = assetSpecs.moveNext()) {
				double start = 0;
				double end = 1;
				if ( mustUseMeasure ) {
					start = attr.getDouble("STARTMEASURE");
					end = attr.getDouble("ENDMEASURE");
				}
				segments.add(new LinearSegment( start, end, attr.getString("ASSETATTRID"), attr.getDouble("NUMVALUE")));
			}
			assetSpecs.close();
		}
		// Loading features parameter values
		if ( !script.getFeaturesNames().isEmpty() ) {
			sqfWhere = new StringBuilder("(refobjectid, classstructureid) IN (select assetfeatureid,classstructureid from assetfeature where assetnum = :assetnum and siteid = :siteid ) and refobjectname = 'ASSETFEATURE' ");
			if ( mustUseMeasure ) {
				sqfWhere.append(" and endmeasure >= :1 and startmeasure <= :2");
			}
			sqfWhere.append(" and (feature,assetattrid) in ( ");
			for(String featureName : script.getFeaturesNames()) {
				for(String featureParamName : script.getFeatureParamNames(featureName)) {
					sqfWhere.append("('").append(featureName.toUpperCase()).append("','").append(featureParamName).append("'),");
					
				}
			}
			sqfWhere.deleteCharAt(sqfWhere.length()-1);
			sqfWhere.append(')');
			
			sqf = new SqlFormat(asset, sqfWhere.toString());
			if ( mustUseMeasure ) {
				sqf.setDouble(1, woStartMeasure);
				sqf.setDouble(2, woEndMeasure);
			}
			
			MboSetRemote featuresAttrSet = asset.getMboSet("$FEATUREATTR","ASSETFEATURESPEC",sqf.format());
			
			featuresAttrSet.setOrderBy("STARTMEASURE, ENDMEASURE asc");
			for(MboRemote featureAttr = featuresAttrSet.moveFirst(); featureAttr != null; featureAttr = featuresAttrSet.moveNext()) {
				double start = 0;
				double end = 1;
				if ( mustUseMeasure ) {
					start = featureAttr.getDouble("STARTMEASURE");
					end = featureAttr.getDouble("ENDMEASURE");
				}
				segments.add(new LinearSegment(start, end, featureAttr.getString("FEATURE") + ":" + featureAttr.getString("ASSETATTRID"), featureAttr.getDouble("NUMVALUE")));
			}
			featuresAttrSet.close();
		}
		segments = LinearSegment.aggregate(segments);
//		Collections.sort(segments);
		
		return segments;
	}
	 */
}
