package com.ibm.swg.app.jobplan;

import java.rmi.RemoteException;

import com.ibm.swg.util.DJPConstants;
import com.ibm.swg.util.DynamicJobPlanHelper;

import psdi.mbo.Mbo;
import psdi.mbo.MboValue;
import psdi.mbo.StatefulMboRemote;
import psdi.mbo.SynonymDomain;
import psdi.mbo.Translate;
import psdi.util.MXException;

public class FldJobCalcMethod extends SynonymDomain implements DJPConstants {
	
	
	public FldJobCalcMethod(MboValue mbv) {
		super(mbv);
		setDomainId("CALCMETHOD");
		
		try {
			String objectName = mbv.getMbo().getThisMboSet().getMboSetInfo().getObjectName();
			Translate t = mbv.getMbo().getTranslator();
			if ( objectName.equals("JOBTASK")) {
				setConditionalValWhere("type in (" + t.toExternalList("CALCAPPLYTO", "HOURS") + ")");
			} else if ( objectName.equals("WPMATERIAL")) {
				setConditionalValWhere("type in (" + t.toExternalList("CALCAPPLYTO", "QUANTITY") + ")");
			} 
		} catch (RemoteException e) {
			e.printStackTrace();
		} catch (MXException e) {
			e.printStackTrace();
		}
		
	}
	
	
	@Override
	public void init() throws MXException, RemoteException {
		super.init();
		MboValue mbv = getMboValue();
		Mbo mbo = mbv.getMbo();	// could be a JOBTASK, JLOBLABOR, JOBMATERIAL, JOBSERVICE, JOBTOOL, WPLABOR, WOACTIVITY, WPLABOR, WPMATERIAL, WPSERVICE, WPTOOL 
		StatefulMboRemote parentMbo;
		
		DynamicJobPlanHelper.onJobResourceMethodchanged(mbo);
		if ( !mbo.isBasedOn("JOBPLAN") && mbo instanceof StatefulMboRemote) {
			parentMbo = (StatefulMboRemote) mbo.getOwner();	// could be a WORKORDER or JOBPLAN
		} else if ( mbo instanceof StatefulMboRemote ) {
			parentMbo = (StatefulMboRemote) mbo;
		} else return;
		 
		
		if ( parentMbo == null ) return;
		
		boolean cond = parentMbo.isBasedOn("JOBPLAN") && parentMbo.getString("STATUS").equals("ACTIVE");
//		cond |= parentMbo.isBasedOn("WORKORDER") && !parentMbo.getInternalStatus().equals("WAPPR");
//		cond |= parentMbo.isBasedOn("WOACTIVITY") && !parentMbo.getInternalStatus().equals("WAPPR");
		
		mbv.setFlag(READONLY, cond);
		mbv.setFlag(REQUIRED, true);
		
	}

	@Override
	public void action() throws MXException, RemoteException {
	
		String[] fieldToActivate;
		String[] fieldToDeactivate;
		if ( getMboValue().getString().equals(CALCMETHOD_LEVEL)) {
			fieldToActivate = new String[]{CALCAPPLYTO, CALCLEVEL};
			fieldToDeactivate = new String[]{CALCEXPRESSION};//CALCATTRIBUTE};
		} else if ( getMboValue().getString().equals(CALCMETHOD_PROPORTIONAL)) {
			fieldToActivate = new String[]{CALCAPPLYTO};
			fieldToDeactivate = new String[]{CALCLEVEL,CALCEXPRESSION};//CALCATTRIBUTE};
		} else if ( getMboValue().getString().equals(CALCMETHOD_PROPORTIONAL_ATTRIBUTE)) {
			fieldToActivate = new String[]{CALCAPPLYTO, CALCEXPRESSION};//CALCATTRIBUTE};
			fieldToDeactivate = new String[]{CALCLEVEL};
		} else if ( getMboValue().getString().equals(CALCMETHOD_LEVEL_ATTRIBUTE)) {
			fieldToActivate = new String[]{CALCAPPLYTO, CALCLEVEL, CALCEXPRESSION};//CALCATTRIBUTE};
			fieldToDeactivate = new String[]{};
		} else if ( getMboValue().getString().equals(CALCMETHOD_STATIC)) {
			fieldToActivate = new String[]{};
			fieldToDeactivate = new String[]{CALCLEVEL, CALCEXPRESSION, CALCAPPLYTO};
		} else {
			fieldToActivate = new String[]{CALCAPPLYTO};
			fieldToDeactivate = new String[]{CALCLEVEL, CALCEXPRESSION};//CALCATTRIBUTE};
		}
		
		for(String fieldName : fieldToActivate) {
			MboValue mbv = getMboValue(fieldName);
			mbv.setFlag(READONLY, false);	
			mbv.setFlag(REQUIRED, true);
			mbv.setValueNull(NOACCESSCHECK | NOACTION | NOVALIDATION);
		}
		
		for(String fieldName : fieldToDeactivate) {
			MboValue mbv = getMboValue(fieldName);
			mbv.setFlag(READONLY, true);
			mbv.setFlag(REQUIRED, false);
			mbv.setValueNull(NOACCESSCHECK | NOACTION | NOVALIDATION);
		}

		DynamicJobPlanHelper.onJobResourceMethodchanged(mboValue.getMbo());
	}
}
