package com.ibm.swg.app.jobplan;

import java.rmi.RemoteException;

import com.ibm.swg.util.DJPConstants;

import psdi.mbo.MAXTableDomain;
import psdi.mbo.Mbo;
import psdi.mbo.MboSetInfo;
import psdi.mbo.MboValue;
import psdi.mbo.StatefulMboRemote;
import psdi.mbo.Translate;
import psdi.util.MXException;

public class FldJobCalcLevel extends MAXTableDomain {

	public FldJobCalcLevel(MboValue mbv) {
		super(mbv);

		try {
			MboSetInfo info = mbv.getMbo().getThisMboSet().getMboSetInfo();
			String mboObjectName = info.getObjectName();
			Translate t = mbv.getMbo().getTranslator();
			String relation;
			String listWhere;
			if ( mboObjectName.equals("JOBTASK") ) {
				listWhere = " type in (" + t.toExternalList("CALCAPPLYTO", "HOURS") + ")";
				relation = "resourcelevelsnum = :calclevel and " + listWhere;
			} else if ( mboObjectName.equals("JOBMATERIAL") ) {
				listWhere = " type in (" + t.toExternalList("CALCAPPLYTO", "QUANTITY") + ")";
				relation = "resourcelevelsnum = :calclevel and " + listWhere;
			} else {
				listWhere = " type = :calcapplyto";
				relation = "resourcelevelsnum = :calclevel and " + listWhere;
			}
			setRelationship("RESOURCELEVELS", relation);
			setListCriteria(listWhere);
		} catch (RemoteException e) {
			e.printStackTrace();
		} catch (MXException e) {
			e.printStackTrace();
		}
		
		
		String[] targetKeys = {"calclevel"};
		String[] sourceKeys = {"resourcelevelsnum"};
		setLookupKeyMapInOrder(targetKeys, sourceKeys);
	}
	
	@Override
	public void init() throws MXException, RemoteException {
		super.init();
		
		MboValue mbv = getMboValue();
		Mbo mbo = mbv.getMbo();	// could be a JOBTASK, JLOBLABOR, JOBMATERIAL, JOBSERVICE, JOBTOOL, WPLABOR, WOACTIVITY, WPLABOR, WPMATERIAL, WPSERVICE, WPTOOL 
		StatefulMboRemote parentMbo;
		
		if ( !mbo.isBasedOn("JOBPLAN") && mbo instanceof StatefulMboRemote) {
			parentMbo = (StatefulMboRemote) mbo.getOwner().getOwner();	// could be a WORKORDER or JOBPLAN
		} else if ( mbo instanceof StatefulMboRemote ) {
			parentMbo = (StatefulMboRemote) mbo;
		} else return;
		 
		
		if ( !mbv.isNull() && !getMboValue(DJPConstants.CALCMETHOD).getString().equals(DJPConstants.CALCMETHOD_LEVEL) && !getMboValue(DJPConstants.CALCMETHOD).getString().equals(DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE) ) {
			mbv.setValueNull(NOACCESSCHECK);
		}
		
		if ( parentMbo == null ) return;
		
		boolean cond = parentMbo.isBasedOn("JOBPLAN") && parentMbo.getString("STATUS").equals("ACTIVE");
//		cond |= parentMbo.isBasedOn("WORKORDER") && !parentMbo.getInternalStatus().equals("WAPPR");
//		cond |= parentMbo.isBasedOn("WOACTIVITY") && !parentMbo.getInternalStatus().equals("WAPPR");
		cond |= !(getMboValue(DJPConstants.CALCMETHOD).getString().equals(DJPConstants.CALCMETHOD_LEVEL) || 
				getMboValue(DJPConstants.CALCMETHOD).getString().equals(DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE));

		mbv.setFlag(READONLY, cond);
		
		boolean required = getMboValue(DJPConstants.CALCMETHOD).getString().equals(DJPConstants.CALCMETHOD_LEVEL) ||
				getMboValue(DJPConstants.CALCMETHOD).getString().equals(DJPConstants.CALCMETHOD_LEVEL_ATTRIBUTE);
		mbv.setFlag(REQUIRED, required);
		
	}

}
