package com.ibm.swg.app.resourcelevels;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.List;

import psdi.mbo.Mbo;
import psdi.mbo.MboServerInterface;
import psdi.mbo.MboSet;
import psdi.util.MXApplicationException;
import psdi.util.MXException;

public class ResourceByLevelSet extends MboSet implements ResourceByLevelSetRemote {

	public ResourceByLevelSet(MboServerInterface ms) throws RemoteException {
		super(ms);
	}

	@Override
	protected Mbo getMboInstance(MboSet ms) throws MXException, RemoteException {
		return new ResourceByLevel(ms);
	}

	@Override
	public 	ResourceByLevelRemote getLevelForValue(double value) throws RemoteException, MXException {
		reset();
		resetQbe();

		setWhere(value + " > " + ResourceByLevelRemote.LOWERTHRESHOLD + " AND " + value + " <= "+ ResourceByLevelRemote.UPPERTHRESHOLD);
		
		return (ResourceByLevelRemote) getMbo(0);
		
	}
	
	private static class Segment {
		public double start;
		public double end;
		
		public Segment(double start, double end) {
			this.start = start;
			this.end = end;
		}
		
		public boolean intersect(Segment s) {
			return this.start > s.start && this.start < s.end || this.end > s.start && this.end < s.end;
		}
	}
	
	@Override
	public void validate() throws MXException, RemoteException {
		super.validate();
		List<Segment> segments = new ArrayList<Segment>();
		for(ResourceByLevelRemote mbo = (ResourceByLevelRemote) moveFirst(); mbo != null; mbo = (ResourceByLevelRemote) moveNext()) {
			segments.add(new Segment(mbo.getLowerThreshold(), mbo.getUpperThreshold()));
		}
		
		Segment head = segments.isEmpty() ? null : segments.remove(0);
		while ( head != null) {
			for(Segment tmp : segments) {
				if ( head.intersect(tmp)) {
					throw new MXApplicationException("dynamicjobplan","ThresholdsInvalid");
				}
			}
			
			if ( !segments.isEmpty() ) {
				head = segments.remove(0);
			} else {
				head = null;
			}
		};
	}
	
}
